"""
Hyperparameter Distributions
====================================
Here you'll find a few hyperparameter distributions. It's also possible to create yours by inheriting
from the base class. Each distribution must override the method ``rvs``, which will return a sampled value from
the distribution.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import math
import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Callable, List, Optional

import numpy as np
from scipy.integrate import quad
from scipy.stats import norm, truncnorm


HPSampledValue = Any


class HyperparameterDistribution(metaclass=ABCMeta):
    """
    Base class for other hyperparameter distributions.

    This is the abstract class of the abstract classes.
    Please inherit from these mid-level abstract classes:

    - :class:`~neuraxle.hyperparams.distributions.ContinuousHyperparameterDistribution`
    - :class:`~neuraxle.hyperparams.distributions.DiscreteHyperparameterDistribution`
    - :class:`~neuraxle.hyperparams.distributions.OrdinalHyperparameterDistribution`
    """

    def __init__(self, null_default_value: Any = None):
        """
        Create a HyperparameterDistribution. This method should still be called with super if it gets overriden.
        """
        self.first_id = id(self)
        self.null_default_value = null_default_value

    @abstractmethod
    def is_discrete(self) -> bool:
        raise NotImplementedError("Please use a concrete class.")

    @abstractmethod
    def rvs(self) -> HPSampledValue:
        """
        Sample the random variable.

        :return: The randomly sampled value.
        """
        pass

    def rvs_many(self, size: int) -> List[HPSampledValue]:
        return [self.rvs() for _ in range(size)]

    def nullify(self) -> HPSampledValue:
        return self.null_default_value

    @abstractmethod
    def pdf(self, x: HPSampledValue) -> float:
        """
        Abstract method for probability distribution function value at `x`.

        :param x: value where the probability distribution function is evaluated.

        :return: The probability distribution function value.
        """
        pass

    @abstractmethod
    def cdf(self, x: HPSampledValue) -> float:
        """
        Abstract method for cumulative distribution function value at `x`.
        Therefore, `p = cdf(x)`.

        :param x: value where the cumulative distribution function is evaluated.

        :return: The cumulative distribution function value.
        """
        pass

    @abstractmethod
    def icdf(self, p: float) -> HPSampledValue:
        """
        This is the inverse of the cumulative distribution function,
        where x = icdf(p).

        :param x: value where the cumulative distribution function is evaluated.

        :return: The cumulative distribution function value.
        """
        pass

    @abstractmethod
    def min(self) -> float:
        """
        Abstract method for obtaining minimum value that can sampled in distribution.

        :return: minimal value that can be sampled from distribution.
        """
        pass

    @abstractmethod
    def max(self) -> float:
        """
        Abstract method for obtaining maximal value that can sampled in distribution.

        :return: maximal value that can be sampled from distribution.
        """
        pass

    @abstractmethod
    def mean(self) -> float:
        """
        Abstract method for calculating distribution mean value.

        :return: distribution mean value.
        """
        pass

    def std(self) -> float:
        """
        Base method to calculate distribution std value by taking sqrt of variance value.

        :return: distribution std value.
        """
        return math.sqrt(self.var())

    @abstractmethod
    def var(self) -> float:
        """
        Abstract method for calculate distribution variance value.

        : return: distribution variance value.
        """
        pass

    def __eq__(self, other):
        return self.first_id == other.first_id

    def __contains__(self, item) -> bool:
        """
        Check if item is in the distribution.
        The item is expected to be the result of a call to :func:`HyperparameterDistribution.rvs`.
        """
        return (
            item == self.null_default_value or
            (self.min() <= item <= self.max())
        )

    def __str__(self) -> str:
        return self.__class__.__name__ + f"<min={self.min()}, max={self.max()}>"


class ContinuousHyperparameterDistribution(HyperparameterDistribution):
    """
    Continuous distributions.

    .. seealso::
        :class:`~neuraxle.hyperparams.distributions.DiscreteHyperparameterDistribution`
        :class:`~neuraxle.hyperparams.distributions.OrdinalHyperparameterDistribution`
    """

    def __init__(self, null_default_value=None):
        HyperparameterDistribution.__init__(self, null_default_value=null_default_value)

    def is_discrete(self) -> bool:
        return False

    def icdf(self, p: float) -> HPSampledValue:
        """
        This is the inverse of the cumulative distribution function,
        where x = icdf(p).
        """
        _tolerance = 2e-11
        _tolerance *= (self._pseudo_max() - self._pseudo_min())  # tolerance is relative to range.
        x: HPSampledValue = self._icdf(p, self._pseudo_min(), self._pseudo_max(), _tolerance=2e-11)
        return x  # x is float since it's continuous.

    def _icdf(self, p: float, x_min: float, x_max: float, _tolerance=2e-11, _maxiters=100) -> float:

        if p <= 0.0:
            return x_min
        elif p >= 1.0:
            return x_max

        for _ in range(0, max(_maxiters, 1)):

            # Calculating the mean of the current interval
            x_mid = (x_min + x_max) / 2.0
            p_mid = self.cdf(x_mid)
            p_diff = p - p_mid

            # Exit condition
            if abs(p_diff) < _tolerance:
                return x_mid

            # Update rule for next iter: if p is greater than p_mid: increase x_min. Otherwise decrease x_max.
            if p_diff > 0:
                x_min = x_mid
            else:
                x_max = x_mid

        warnings.warn(
            f"Could not find a solution for icdf({p}): reached at most `{x_min} < x < {x_max}`. "
            f"returning midpoint: {x_mid}. ")
        return x_mid

    def _pseudo_min(self) -> float:
        """
        This is like the function min, but in case it is infinite, there will be a clipping to at least 4 sigmas.
        """
        x = self.min()
        if math.isinf(x) or math.isnan(x):
            x = self.mean() - 4 * self.std()
        return x

    def _pseudo_max(self) -> float:
        """
        This is like the function max, but in case it is infinite, there will be a clipping to at most 4 sigmas.
        """
        x = self.max()
        if math.isinf(x) or math.isnan(x):
            x = self.mean() + 4 * self.std()
        return x


class DiscreteHyperparameterDistribution(HyperparameterDistribution):
    """
    Discrete distributions.

    Ineritors will have to implement the following method at least:
    :func:`DiscreteHyperparameterDistribution.values`.

    .. seealso::
        :class:`~neuraxle.hyperparams.distributions.ContinuousHyperparameterDistribution`
        :class:`~neuraxle.hyperparams.distributions.OrdinalHyperparameterDistribution`
    """

    def __init__(self, null_default_value=None):
        HyperparameterDistribution.__init__(self, null_default_value=null_default_value)

    def is_discrete(self) -> bool:
        return True

    @abstractmethod
    def values(self) -> List[Any]:
        return [i for i in range(int(self.min()), int(self.max()) + 1)]

    def icdf(self, p: float) -> HPSampledValue:
        """
        This is the inverse of the cumulative distribution function,
        where x = icdf(p).
        """
        probas = np.array([self.cdf(i) for i in self.values()])
        closest_idx: int = np.argmin(np.abs(probas - p))
        x: HPSampledValue = self.values()[closest_idx]
        return x

    def probabilities(self) -> List[float]:
        return [self.pdf(i) for i in self.values()]

    def __contains__(self, item) -> bool:
        return item == self.null_default_value or item in self.values()


class OrdinalDiscreteHyperparameterDistribution(DiscreteHyperparameterDistribution):
    """
    Inherit from this class to represent a :class:`DiscreteHyperparameterDistribution`
    that contains values in :func:`DiscreteHyperparameterDistribution.values` that are
    in order such that they would sort the same way, either ascending or descending.

    The usage of this class indicates the AutoML wheter it should consider the different
    values like a categorical or a numerical variable.

    .. seealso::
        :class:`~neuraxle.hyperparams.distributions.ContinuousHyperparameterDistribution`
        :class:`~neuraxle.hyperparams.distributions.DiscreteHyperparameterDistribution`
    """

    def __init__(self, null_default_value=None):
        DiscreteHyperparameterDistribution.__init__(self, null_default_value=null_default_value)


class LogSpaceDistributionMixin:
    """
    Use this mixin when your distribution samples from a log-space distribution
    to identify it.
    """

    def __contains__(self, item) -> bool:
        return item != 0.0 and super().__contains__(item)


class FixedHyperparameter(DiscreteHyperparameterDistribution):
    """This is an hyperparameter that won't change again, but that is still expressed as a distribution."""

    def __init__(self, value, null_default_value=None):
        """
        Create a still hyperparameter

        :param value: what will be returned by calling ``.rvs()``.
        """
        DiscreteHyperparameterDistribution.__init__(self, null_default_value)
        self.value = value

    def rvs(self):
        """
        Sample the non-random anymore value.

        :return: the value given at creation.
        """
        return self.value

    def pdf(self, x) -> float:
        """
        Probability distribution function value at `x`.
        Since the parameter is fixed, the value return is 1 when x == value and 0 otherwise.

        :param x: value where the probability distribution function is evaluated.

        :return: The probability distribution function value.
        """
        if x == self.value:
            return 1.
        return 0.

    def cdf(self, x) -> float:
        """
        Cumulative distribution function value at `x`.
        Since the parameter is fixed, the value return is 1 if x>= value and 0 otherwise.

        :param x: value where the cumulative distribution function is evaluated.

        :return: The cumulative distribution function value.
        """
        if x >= self.value:
            return 1.

        return 0.

    def min(self):
        """
        Calculate minimum value that can be sampled in a fixed distribution.

        :return: minimal value return from distribution.
        """
        return self.value

    def max(self):
        """
        Calculate maximum value that can be sampled in a fixed distribution.

        :return: maximum value return from distribution.
        """
        return self.value

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        return self.value

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        return 0

    def __str__(self):
        return self.__class__.__name__ + f"(value={self.value}: {type(self.value)})"

    def values(self) -> List[Any]:
        return [self.value]


# TODO: Mixin this or something:
# class DelayedAdditionOf(MalleableDistribution):
#     """A HyperparameterDistribution (MalleableDistribution mixin) that """
#
#     def __init__(self, *dists):
#         self.dists = dists
#
#     def rvs(self):
#         rvss = [d.rvs if hasattr(d, 'rvs') else d for d in self.dists]
#         return sum(rvss)
#
#
# class MalleableDistribution(metaclass=ABCMeta):
#     """An hyperparameter distribution to which it's possible to do additional math using defaut python operators."""
#
#     def __add__(self, other):
#         return DelayedAdditionOf(self, other)
#
# max min + - / * % ** // == != < > <= >=
#


class Boolean(DiscreteHyperparameterDistribution):
    """Get a random boolean hyperparameter."""

    def __init__(self, proba_is_true: Optional[float] = None, null_default_value=False):
        """
        Create a boolean hyperparameter with given probability.

        Boolean distribution is in fact a Bernouilli distribution where given probability
        set occurance probability of having value 1. (1 - probability) gives occurance probability of having value 0.

        :param proba: a float corresponding to proportion of 1 over 0.
        :type proba: float
        :param null_default_value: default value for distribution
        :type null_default_value: default choice value. if None, default choice value will be the first choice
        """
        DiscreteHyperparameterDistribution.__init__(self, null_default_value)

        if proba_is_true is None:
            proba_is_true = 0.5

        if not (0 <= proba_is_true <= 1):
            raise ValueError("Probability must be between 0 and 1 (inclusively).")

        self.proba_is_true = proba_is_true

    def probabilities(self) -> List[float]:
        return [self.proba_is_true, 1 - self.proba_is_true]

    def values(self):
        return [True, False]

    def rvs(self) -> bool:
        """
        Get a random True or False.

        :return: True or False (random).
        """
        return bool(np.random.choice([False, True], p=[1 - self.proba_is_true, self.proba_is_true]))

    def pdf(self, x) -> float:
        """
        Calculate the boolean probability mass function value at position `x`.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        if (x is True) or (x == 1):
            return self.proba_is_true

        if (x is False) or (x == 0):
            return 1 - self.proba_is_true

        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the boolean cumulative distribution function value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        if x < 0:
            return 0.

        if (0 <= x < 1) or (x is False):
            return 1 - self.proba_is_true

        if x >= 1 or (x is True):
            return 1.

        return 0.

    def min(self):
        """
        Calculate minimum value that can be sampled in a boolean distribution.

        :return: minimal value return from distribution.
        """
        return 0

    def max(self):
        """
        Calculate maximum value that can be sampled in a boolean distribution.

        :return: maximum value return from distribution.
        """
        return 1

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        return self.proba_is_true

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        return self.proba_is_true * (1 - self.proba_is_true)

    def __repr__(self):
        return self.__class__.__name__ + f"(proba_is_true={self.proba_is_true})"


class Choice(DiscreteHyperparameterDistribution):
    """Get a random value from a choice list of possible value for this hyperparameter.

    When narrowed, the choice will only collapse to a single element when narrowed enough.
    For example, if there are 4 items in the list, only at a narrowing value of 0.25 that
    the first item will be kept alone.
    """

    def __init__(self, choice_list: List[Any], probas: Optional[List[float]] = None, null_default_value=None):
        """
        Create a random choice hyperparameter from the given list.

        :param choice_list: a list of values to sample from.
        :type choice_list: List
        :param null_default_value: default value for distribution
        :type null_default_value: default choice value. if None, default choice value will be the first choice
        """
        DiscreteHyperparameterDistribution.__init__(self, null_default_value)

        self.choice_list: List[Any] = choice_list

        if probas is None:
            probas = [1 / len(self.choice_list) for _ in self.choice_list]

        # Normalize probas juste in case sum is not equal to one.
        self.probas: List[float] = (np.array(probas) / np.sum(probas)).tolist()

    def probabilities(self) -> List[float]:
        return self.probas

    def values(self) -> List[Any]:
        return self.choice_list

    def rvs(self):
        """
        Get one of the items randomly.

        :return: one of the items of the list.
        """
        choice_index = np.arange(0, len(self), 1)
        rvs_index = np.random.choice(choice_index, p=self.probas)
        return self.choice_list[rvs_index]

    def pdf(self, x) -> float:
        """
        Calculate the choice probability mass function value at position `x`.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        try:
            x_in_choice = x in self.choice_list
        except (TypeError, ValueError, AttributeError):
            raise ValueError(
                "Item not find in list. Make sure the item is in the choice list and a correct method  __eq__ is defined for all item in the list.")
        else:
            if x_in_choice:
                index = get_index_in_list_with_bool(self.choice_list, x)
                return self.probas[index]

        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the choice probability cumulative distribution function value at position `x`.
        The index in the list is used to know how the choice is performed.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        try:
            index = get_index_in_list_with_bool(self.choice_list, x)
        except ValueError:
            raise ValueError(
                "Item not found in list. Make sure the item is in the choice list and a correct method  __eq__ is defined for all item in the list.")
        except (NotImplementedError, NotImplemented):
            raise ValueError("A correct method for __eq__ should be defined for all item in the list.")
        except AttributeError:
            raise ValueError("choice_list param should be a list.")
        else:
            probas = np.array(self.probas)
            return np.sum(probas[0:index + 1])

    def __len__(self):
        """
        Return the number of choices.

        :return: the number of choices.
        """
        return len(self.choice_list)

    def min(self):
        """
        Calculate minimum value that can be sampled in a choice distribution.

        Here the minimal index in the list is return. In this case it returns 0.

        :return: minimal value return from distribution.
        """
        return 0.

    def max(self):
        """
        Calculate maximal value that can be sampled in a choice distribution.

        Here the maximal index in the list is return. In this case it returns the length value of the list.

        :return: maximal value return from distribution.
        """
        return len(self) - 1

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        return mean

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        second_moment = np.sum(choice_index**2 * probas)
        var = second_moment - mean**2
        return var

    def __str__(self):
        return self.__class__.__name__ + \
            f"({'choice_list='+', '.join(self.choice_list) if len(self.choice_list) < 4 else f'<a {len(self.choice_list)} elements list>'})"


class PriorityChoice(OrdinalDiscreteHyperparameterDistribution):
    """
    Get a random value from a choice list of possible value for this hyperparameter.

    This is a choice that is ordinal and where the first choice is the most important one, as well as the default.

    The first parameters are kept until the end when the list is narrowed (it is narrowed progressively),
    unless there is a best guess that surpasses some of the top choices.
    """

    def __init__(self, choice_list: List[Any], probas: Optional[List[float]] = None, null_default_value=None):
        """
        Create a random choice hyperparameter from the given list (choice_list).
        The first parameters in the choice_list will be kept longer when narrowing the space.

        :param choice_list: a list of values to sample from.
        :type choice_list: List
        :param null_default_value: default value for distribution
        :type null_default_value: default choice value. if None, default choice value will be the first choice
        """
        if null_default_value is None:
            null_default_value = choice_list[0]
        elif null_default_value in choice_list:
            null_default_value = null_default_value
        else:
            raise ValueError(
                'invalid default value {0} not in choice list : {1}'.format(null_default_value, choice_list))

        OrdinalDiscreteHyperparameterDistribution.__init__(self, null_default_value)
        self.choice_list: List[Any] = choice_list

        if probas is None:
            probas = [1 / len(self.choice_list) for _ in self.choice_list]

        # Normalize probas juste in case sum is not equal to one.
        self.probas: List[float] = list(np.array(probas) / np.sum(probas))

    def probabilities(self) -> List[float]:
        return self.probas

    def values(self) -> List:
        return self.choice_list

    def rvs(self):
        """
        Get one of the items randomly.

        :return: one of the items of the list.
        """
        choice_index = np.arange(0, len(self), 1)
        rvs_index = np.random.choice(choice_index, p=self.probas)
        return self.choice_list[rvs_index]

    def pdf(self, x) -> float:
        """
        Calculate the choice probability mass function value at position `x`.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        try:
            x_in_choice = x in self.choice_list
        except (TypeError, ValueError, AttributeError):
            raise ValueError(
                "Item not find in list. Make sure the item is in the choice list and a correct method  __eq__ is defined for all item in the list.")
        else:
            if x_in_choice:
                index = get_index_in_list_with_bool(self.choice_list, x)
                return self.probas[index]

        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the choice probability cumulative distribution function value at position `x`.
        The index in the list is used to know how the choice is performed.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        try:

            index = get_index_in_list_with_bool(self.choice_list, x)
        except ValueError:
            raise ValueError(
                "Item not find in list. Make sure the item is in the choice list and a correct method  __eq__ is defined for all item in the list.")
        except (NotImplementedError, NotImplemented):
            raise ValueError("A correct method for __eq__ should be defined for all item in the list.")
        except AttributeError:
            raise ValueError("choice_list param should be a list.")
        else:
            return np.sum(self.probas[0:index + 1])

    def __len__(self):
        """
        Return the number of choices.

        :return: the number of choices.
        """
        return len(self.choice_list)

    def min(self):
        """
        Calculate minimum value that can be sampled in a priority choice distribution.

        Here the minimal index in the list is return. In this case it returns 0.

        :return: minimal value return from distribution.
        """
        return 0.

    def max(self):
        """
        Calculate maximal value that can be sampled in a priority choice distribution.

        Here the maximal index in the list is return. In this case it returns the length value of the list.

        :return: maximal value return from distribution.
        """
        return len(self) - 1

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        return mean

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        second_moment = np.sum(choice_index ** 2 * probas)
        var = second_moment - mean ** 2
        return var


class WrappedDiscreteHyperparameterDistribution(DiscreteHyperparameterDistribution):
    def __init__(self, hd: HyperparameterDistribution = None, hds: List[HyperparameterDistribution] = None,
                 null_default_value=None):
        """
        Create a wrapper that will surround another HyperparameterDistribution.
        The wrapper might use one (hd) and/or many (hds) HyperparameterDistribution depending on the argument(s) used.

        :param hd: the other HyperparameterDistribution to wrap.
        :param hds: the others HyperparameterDistribution to wrap.
        """
        DiscreteHyperparameterDistribution.__init__(self, null_default_value)
        self.hd: HyperparameterDistribution = hd

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.hd) + ")"

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.hd) + ")"


class Quantized(WrappedDiscreteHyperparameterDistribution):
    """
    A quantized wrapper for another distribution: will round() the returned :func:`HyperparameterDistribution.rvs` number.
    """

    def rvs(self) -> int:
        """
        Will return an integer, rounded from the output of the previous distribution.

        :return: an integer.
        """
        return round(self.hd.rvs())

    def pdf(self, x) -> float:
        """
        Calculate the Quantized probability mass function value at position `x` of a continuous distribution.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        # In order to calculate the pdf for any quantized distribution,
        # we have to perform the integral from x-0.5 to x+0.5 (because of round).
        if isinstance(x, int) or (isinstance(x, float) and x.is_integer()):
            return quad(self.hd.pdf, x - 0.5, x + 0.5)[0]
        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the Quantized cumulative distribution function at position `x` of a continuous distribution.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        # In order to calculate the cdf for any quantized distribution, we have to take the cdf at x + 0.5.
        return self.hd.cdf(math.floor(x) + 0.5)

    def min(self):
        """
        Calculate minimum value that can be sampled in the quanitzed version of the distribution.

        :return: minimal value return from distribution.
        """
        hd_min = self.hd.min()
        if np.isneginf(hd_min):
            return hd_min
        return round(hd_min)

    def max(self):
        """
        Calculate maximal value that can be sampled in the quantized version of the distribution.

        :return: maximal value return from distribution.
        """
        hd_max = self.hd.max()
        if np.isposinf(hd_max):
            return hd_max
        return round(hd_max)

    def mean(self) -> float:
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        # We will perform two sum starting from floor(mean of original distribution) and from ceil(mean of orignal  distribution).
        original_mean = self.hd.mean()

        # First sum
        starting_decreasing_value = math.floor(original_mean)
        decreasing_sum = _calculate_sum(lambda x: x * self.pdf(x), [starting_decreasing_value, -np.inf])

        # Second sum
        starting_increasing_value = math.ceil(original_mean)
        if (starting_increasing_value - starting_decreasing_value) < 1e-10:
            starting_increasing_value += 1

        increasing_sum = _calculate_sum(lambda x: x * self.pdf(x), [starting_increasing_value, np.inf])

        return decreasing_sum + increasing_sum

    def var(self) -> float:
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        # first moment
        first_moment = self.mean()

        # Calculate second moment
        # We will perform two sum starting from floor(mean of original distribution) and from ceil(mean of orignal  distribution).
        original_mean = self.hd.mean()
        # First sum
        starting_decreasing_value = math.floor(original_mean)
        decreasing_sum = _calculate_sum(lambda x: x ** 2 * self.pdf(x), [starting_decreasing_value, -np.inf])
        # Second sum
        starting_increasing_value = math.ceil(original_mean)
        if (starting_increasing_value - starting_decreasing_value) < 1e-10:
            starting_increasing_value += 1
        increasing_sum = _calculate_sum(lambda x: x ** 2 * self.pdf(x), [starting_increasing_value, np.inf])

        second_moment = decreasing_sum + increasing_sum

        return second_moment - first_moment ** 2

    def values(self) -> List[Any]:
        return [i for i in range(int(self.min()), int(self.max()) + 1)]


class RandInt(DiscreteHyperparameterDistribution):
    """Get a random integer within a range"""

    def __init__(self, min_included: int, max_included: int, null_default_value: int = None):
        """
        Create a quantized random uniform distribution.
        A random integer between the two values inclusively will be returned.

        :param min_included: minimum integer, included.
        :param max_included: maximum integer, included.
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            DiscreteHyperparameterDistribution.__init__(self, min_included)
        else:
            DiscreteHyperparameterDistribution.__init__(self, null_default_value)

        self.min_included = min_included
        self.max_included = max_included

    def probabilities(self):
        values = self.values()
        return [1 / len(values) for _ in values]

    def values(self):
        return [i for i in range(self.min_included, self.max_included + 1)]

    def rvs(self) -> int:
        """
        Will return an integer in the specified range as specified at creation.

        :return: an integer.
        """
        return np.random.randint(self.min_included, self.max_included + 1)

    def pdf(self, x) -> float:
        """
        Calculate the random int mass function value at position `x`.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """

        possible_values = set(range(self.min_included, self.max_included + 1))
        if (isinstance(x, int) or x.is_integer()) and x in possible_values:
            return 1 / (self.max_included - self.min_included + 1)

        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the random int cumulative distribution function value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        if x < self.min_included:
            return 0.

        if x > self.max_included:
            return 1.

        return (math.floor(x) - self.min_included + 1) / (self.max_included - self.min_included + 1)

    def min(self):
        """
        Calculate minimum value that can be sampled in the randint distribution.

        :return: minimal value return from distribution.
        """
        return self.min_included

    def max(self):
        """
        Calculate maximal value that can be sampled in the randint distribution.

        :return: maximal value return from distribution.
        """
        return self.max_included

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        return (self.max_included + self.min_included) / 2

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        return ((self.max_included - self.min_included + 1) ** 2 - 1) / 12

    def __str__(self):
        return self.__class__.__name__ + "(" + "max_included=" + str(self.max_included) + ", min_included=" + str(self.min_included) + ")"


class Uniform(ContinuousHyperparameterDistribution):
    """Get a uniform distribution."""

    def __init__(self, min_included: float, max_included: float, null_default_value=None):
        """
        Create a random uniform distribution.
        A random float between the two values somehow inclusively will be returned.

        :param min_included: minimum integer, included.
        :type min_included: float
        :param max_included: maximum integer, might be included - for more info, see `examples <https://docs.python.org/2/library/random.html#random.uniform>`__
        :type max_included: float
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            ContinuousHyperparameterDistribution.__init__(self, min_included)
        else:
            ContinuousHyperparameterDistribution.__init__(self, null_default_value)

        self.min_included: float = min_included
        self.max_included: float = max_included

    def rvs(self) -> float:
        """
        Will return a float value in the specified range as specified at creation.

        :return: a float.
        """
        return np.random.random() * (self.max_included - self.min_included) + self.min_included

    def pdf(self, x):
        """
        Calculate the Uniform probability distribution value at position `x`.

        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """

        if self.min_included == self.max_included and (x == self.min_included):
            return 1.

        if (x >= self.min_included) and (x <= self.max_included):
            return 1 / (self.max_included - self.min_included)

        # Manage case where it is outside the probability distribution function.
        return 0.

    def cdf(self, x):
        """
        Calculate the Uniform cumulative distribution value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """

        if self.min_included == self.max_included and (x == self.min_included):
            return 1.

        if x < self.min_included:
            return 0.

        if (x >= self.min_included) and (x <= self.max_included):
            return (x - self.min_included) / (self.max_included - self.min_included)

        # Manage the case where x_value > self.max_included
        return 1.

    def min(self):
        """
        Calculate minimum value that can be sampled in the uniform distribution.

        :return: minimal value return from distribution.
        """
        return self.min_included

    def max(self):
        """
        Calculate maximal value that can be sampled in the uniform distribution.

        :return: maximal value return from distribution.
        """
        return self.max_included

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        return (self.min_included + self.max_included) / 2

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        return (self.max_included - self.min_included) ** 2 / 12

    def __str__(self):
        return self.__class__.__name__ + "(" + "max_included=" + str(self.max_included) + ", min_included=" + str(self.min_included) + ")"


class LogUniform(LogSpaceDistributionMixin, ContinuousHyperparameterDistribution):
    """Get a LogUniform distribution.

    For example, this is good for neural networks' learning rates: that vary exponentially."""

    def __init__(self, min_included: float, max_included: float, null_default_value=None):
        """
        Create a quantized random log uniform distribution.
        A random float between the two values inclusively will be returned.

        :param min_included: minimum integer, should be somehow included.
        :param max_included: maximum integer, should be somehow included.
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            ContinuousHyperparameterDistribution.__init__(self, math.log2(min_included))
        else:
            ContinuousHyperparameterDistribution.__init__(self, math.log2(null_default_value))

        self.min_included: float = min_included
        self.max_included: float = max_included
        self.log2_min_included = math.log2(min_included)
        self.log2_max_included = math.log2(max_included)

    def rvs(self) -> float:
        """
        Will return a float value in the specified range as specified at creation.

        :return: a float.
        """
        return 2 ** np.random.uniform(self.log2_min_included, self.log2_max_included)

    def pdf(self, x) -> float:
        """
        Calculate the logUniform probability distribution value at position `x`.
        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """
        if (self.log2_min_included == self.log2_max_included) and x == 2 ** self.log2_min_included:
            return 1.

        if (x >= 2 ** self.log2_min_included) and (x <= 2 ** self.log2_max_included):
            return 1 / (x * math.log(2) * (self.log2_max_included - self.log2_min_included))

        return 0.

    def cdf(self, x) -> float:
        """
        Calculate the logUniform cumulative distribution value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        if x < 2 ** self.log2_min_included:
            return 0.

        if (self.log2_min_included == self.log2_max_included) and x == 2 ** self.log2_min_included:
            return 1.

        if (x >= 2 ** self.log2_min_included) and (x <= 2 ** self.log2_max_included):
            return (math.log2(x) - self.log2_min_included) / (self.log2_max_included - self.log2_min_included)

        # Manage the case x > 2**self.log2_max_included
        return 1.

    def min(self):
        """
        Calculate minimum value that can be sampled in the LogUniform distribution.

        :return: minimal value return from distribution.
        """
        return self.min_included

    def max(self):
        """
        Calculate maximal value that can be sampled in the LogUniform distribution.

        :return: maximal value return from distribution.
        """
        return self.max_included

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        if self.log2_min_included == self.log2_max_included:
            return self.log2_max_included

        return (self.max_included - self.min_included) / (
            math.log(2) * (self.log2_max_included - self.log2_min_included))

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        if self.log2_min_included == self.log2_max_included:
            return 0.

        esperance_squared = (self.max_included ** 2 - self.min_included ** 2) / (
            2 * math.log(2) * (self.log2_max_included - self.log2_min_included))
        return esperance_squared - (self.mean() ** 2)

    def __str__(self):
        return self.__class__.__name__ + "(" + "max_included=" + str(self.max_included) + ", min_included=" + str(self.min_included) + ")"


class Normal(ContinuousHyperparameterDistribution):
    """Get a normal distribution."""

    def __init__(self, mean: float, std: float,
                 hard_clip_min: float = None, hard_clip_max: float = None, null_default_value: float = None):
        """
        Create a normal distribution from mean and standard deviation.

        :param mean: the most common value to pop
        :type mean: float
        :param std: the standard deviation (that is, the sqrt of the variance).
        :type std: float
        :param hard_clip_min: if not none, rvs will return max(result, hard_clip_min).
        :type hard_clip_min: float
        :param hard_clip_max: if not none, rvs will return min(result, hard_clip_min).
        :type hard_clip_max: float
        :param null_default_value: if none, null default value will be set to hard_clip_min.
        :type null_default_value: float
        """
        if null_default_value is None:
            ContinuousHyperparameterDistribution.__init__(self, hard_clip_min)
        else:
            ContinuousHyperparameterDistribution.__init__(self, null_default_value)

        self._mean = mean
        self._std = std
        self.hard_clip_min = hard_clip_min
        self.hard_clip_max = hard_clip_max

    def rvs(self) -> float:
        """
        Will return a float value in the specified range as specified at creation.

        :return: a float.
        """
        if self.hard_clip_min is None and self.hard_clip_max is None:
            result = float(np.random.normal(self._mean, self._std))
        else:
            a = -np.inf
            b = np.inf

            if self.hard_clip_min is not None:
                a = (self.hard_clip_min - self._mean) / self._std

            if self.hard_clip_max is not None:
                b = (self.hard_clip_max - self._mean) / self._std

            result = truncnorm.rvs(a=a, b=b, loc=self._mean, scale=self._std)

        if not math.isfinite(result):
            return self.rvs()
        return float(result)

    def pdf(self, x) -> float:
        """
        Calculate the Normal probability distribution value at position `x`.
        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """

        if self.hard_clip_min is not None and (x < self.hard_clip_min):
            return 0.

        if self.hard_clip_max is not None and (x > self.hard_clip_max):
            return 0.

        if self.hard_clip_min is not None or self.hard_clip_max is not None:
            a = -np.inf
            b = np.inf

            if self.hard_clip_min is not None:
                a = (self.hard_clip_min - self._mean) / self._std

            if self.hard_clip_max is not None:
                b = (self.hard_clip_max - self._mean) / self._std

            return truncnorm.pdf(x, a=a, b=b, loc=self._mean, scale=self._std)

        return norm.pdf(x, loc=self._mean, scale=self._std)

    def cdf(self, x) -> float:
        """
        Calculate the Normal cumulative distribution value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        if self.hard_clip_min is not None and (x < self.hard_clip_min):
            return 0.

        if self.hard_clip_max is not None and (x > self.hard_clip_max):
            return 1.

        if self.hard_clip_min is not None or self.hard_clip_max is not None:
            a = -np.inf
            b = np.inf

            if self.hard_clip_min is not None:
                a = (self.hard_clip_min - self._mean) / self._std

            if self.hard_clip_max is not None:
                b = (self.hard_clip_max - self._mean) / self._std

            return truncnorm.cdf(x, a=a, b=b, loc=self._mean, scale=self._std)

        return norm.cdf(x, loc=self._mean, scale=self._std)

    def min(self):
        """
        Calculate minimum value that can be sampled in the Normal distribution.

        :return: minimal value return from distribution.
        """
        return self.hard_clip_min if self.hard_clip_min is not None else -1 * np.inf

    def max(self):
        """
        Calculate minimum value that can be sampled in the Normal distribution.

        :return: minimal value return from distribution.
        """
        return self.hard_clip_max if self.hard_clip_max is not None else np.inf

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        if self.hard_clip_min is None and self.hard_clip_max is None:
            return self._mean

        if self.hard_clip_min is None:
            alpha = -np.inf
        else:
            alpha = (self.hard_clip_min - self._mean) / self._std

        if self.hard_clip_max is None:
            beta = np.inf
        else:
            beta = (self.hard_clip_max - self._mean) / self._std

        Z = norm.cdf(beta) - norm.cdf(alpha)

        mean = self._mean + (norm.pdf(alpha) - norm.pdf(beta)) / Z * self._std
        return mean

    def std(self):
        """
        Calculate standard deviation value of the random variable.

        :return: standard deviation value of the random variable.
        """
        if self.hard_clip_min is None and self.hard_clip_max is None:
            return self._std

        return math.sqrt(self.var())

    def var(self):
        """
        Calculate variance value of the random variable.

        :return: variance value of the random variable.
        """
        if self.hard_clip_min is None and self.hard_clip_max is None:
            return self._std ** 2

        if self.hard_clip_min is None:
            alpha = -np.inf
        else:
            alpha = (self.hard_clip_min - self._mean) / self._std

        if self.hard_clip_max is None:
            beta = np.inf
        else:
            beta = (self.hard_clip_max - self._mean) / self._std

        Z = norm.cdf(beta) - norm.cdf(alpha)

        if self.hard_clip_max is None:
            # Case of one side truncated gaussian (lower tail truncated).
            variance = self._std ** 2 * (1 + alpha * norm.pdf(alpha) / Z - (norm.pdf(alpha) / Z) ** 2)
            return variance

        if self.hard_clip_min is None:
            # Case of one side truncated gaussian (upper tail truncated).
            variance = self._std ** 2 * (1 + - beta * norm.pdf(beta) / Z - (norm.pdf(beta) / Z) ** 2)
            return variance

        # Case of both side truncated gaussian.
        variance = self._std ** 2 * (1 + (alpha * norm.pdf(alpha) - beta * norm.pdf(beta)) / Z - (
            (norm.pdf(alpha) - norm.pdf(beta)) / Z) ** 2)
        return variance

    def __str__(self):
        return self.__class__.__name__ + f"(_mean={str(self._mean)}, _std={str(self._std)}" \
                                         f", hard_clip_min={self.hard_clip_min}, hard_clip_max={self.hard_clip_max})"


class LogNormal(LogSpaceDistributionMixin, ContinuousHyperparameterDistribution):
    """Get a LogNormal distribution."""

    @staticmethod
    def from_regular_mean_std(
        mean: float,
        std: float,
        hard_clip_min: float = None,
        hard_clip_max: float = None
    ):
        # See some theory here::
        # https://en.wikipedia.org/wiki/Log-normal_distribution#Generation_and_parameters
        # https://en.wikipedia.org/wiki/Log-normal_distribution#Arithmetic_moments

        # See these resources for more information on how to compute things:
        # https://www.wolframalpha.com/input?i=x%5E2%3De%5E%282*y%2B2*z%5E2%29-e%5E%282*y%2Bz%5E2%29
        # https://www.wolframalpha.com/input?i=sqrt%28log%281%2F2+e%5E%28-2+y%29+%28e%5E%282+y%29+%2B+sqrt%28e%5E%284+y%29+%2B+4+e%5E%282+y%29+x%5E2%29%29%29%29
        # https://www.wolframalpha.com/input?i=sqrt%28log%281%2F2+e%5E%28-2+y%29+%28e%5E%282+y%29+%2B+sqrt%28e%5E%284+y%29+%2B+4+e%5E%282+y%29+x%5E2%29%29%29%29%2C+y%3D1.6%2C+x%3D+180.601+
        # https://homepage.divms.uiowa.edu/~mbognar/applets/lognormal.html
        # https://stats.stackexchange.com/questions/568768/estimating-the-log-space-sigma-std-parameter-of-a-lognormal-distribution-from
        log2_space_mean = math.log2(mean)
        # log2_space_std = (math.log2(mean - std) + math.log2(mean - std)) / 2

        y = log2_space_mean
        x = std
        # The conversion to base 2 as follow was made just by replacing the log by log2 and the e by 2:
        # sqrt(log(1/2 e^(-2 y) (e^(2 y) + sqrt(e^(4 y) + 4 e^(2 y) x^2))))
        log2_space_std = math.sqrt(math.log2(
            0.5 * 2**(-2 * y) * (2**(2 * y) + math.sqrt(2**(4 * y) + 4 * 2**(2 * y) * x**2))
        ))

        return LogNormal(
            log2_space_mean=log2_space_mean,
            log2_space_std=log2_space_std,
            hard_clip_min=hard_clip_min,
            hard_clip_max=hard_clip_max
        )

    def __init__(self, log2_space_mean: float, log2_space_std: float,
                 hard_clip_min: float = None, hard_clip_max: float = None, null_default_value=None):
        """
        Create a LogNormal distribution.

        :param log2_space_mean: the most common value to pop, but before taking 2**value.
        :param log2_space_std: the standard deviation of the most common value to pop, but before taking 2**value.
        :param hard_clip_min: if not none, rvs will return max(result, hard_clip_min). This value is not checked in logspace (so it is checked after the exp).
        :param hard_clip_max: if not none, rvs will return min(result, hard_clip_min). This value is not checked in logspace (so it is checked after the exp).
        :param null_default_value: null default value for distribution. if None, take the hard_clip_min
        :type null_default_value: int
        """
        if null_default_value is None:
            ContinuousHyperparameterDistribution.__init__(self, hard_clip_min)
        else:
            ContinuousHyperparameterDistribution.__init__(self, null_default_value)
        self.log2_space_mean = log2_space_mean
        self.log2_space_std = log2_space_std
        if hard_clip_min is not None and hard_clip_min <= 0:
            hard_clip_min = None
        self.hard_clip_min = hard_clip_min
        self.hard_clip_max = hard_clip_max

    def rvs(self) -> float:
        """
        Will return a float value in the specified range as specified at creation.
        Note: the range at creation was in log space. The return value is after taking an exponent.

        :return: a float.
        """
        if self.hard_clip_min is None and self.hard_clip_max is None:
            result = 2 ** float(np.random.normal(self.log2_space_mean, self.log2_space_std))
        else:
            a = -np.inf
            b = np.inf

            if self.hard_clip_min is not None:
                a = (math.log2(self.hard_clip_min) - self.log2_space_mean) / self.log2_space_std

            if self.hard_clip_max is not None:
                b = (math.log2(self.hard_clip_max) - self.log2_space_mean) / self.log2_space_std

            result = 2 ** float(truncnorm.rvs(a=a, b=b, loc=self.log2_space_mean, scale=self.log2_space_std))

        if not math.isfinite(result):
            return self.rvs()

        return float(result)

    def pdf(self, x) -> float:
        """
        Calculate the LogNormal probability distribution value at position `x`.
        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """

        if self.hard_clip_min is not None and (x < self.hard_clip_min):
            return 0.

        if self.hard_clip_max is not None and (x > self.hard_clip_max):
            return 0.

        if x <= 0:
            return 0.

        cdf_min = 0.
        cdf_max = 1.

        if self.hard_clip_min is not None:
            cdf_min = norm.cdf(math.log2(self.hard_clip_min), loc=self.log2_space_mean, scale=self.log2_space_std)

        if self.hard_clip_max is not None:
            cdf_max = norm.cdf(math.log2(self.hard_clip_max), loc=self.log2_space_mean, scale=self.log2_space_std)

        pdf_x = 1 / (x * math.log(2) * self.log2_space_std * math.sqrt(2 * math.pi)) * math.exp(
            -(math.log2(x) - self.log2_space_mean) ** 2 / (2 * self.log2_space_std ** 2))
        return pdf_x / (cdf_max - cdf_min)

    def cdf(self, x) -> float:
        """
        Calculate the LogNormal cumulative distribution value at position `x`.
        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        if self.hard_clip_min is not None and (x < self.hard_clip_min):
            return 0.

        if self.hard_clip_max is not None and (x >= self.hard_clip_max):
            return 1.

        if x <= 0:
            return 0.

        cdf_min = 0.
        cdf_max = 1.

        if self.hard_clip_min is not None:
            cdf_min = norm.cdf(math.log2(self.hard_clip_min), loc=self.log2_space_mean, scale=self.log2_space_std)

        if self.hard_clip_max is not None:
            cdf_max = norm.cdf(math.log2(self.hard_clip_max), loc=self.log2_space_mean, scale=self.log2_space_std)

        cdf_x = norm.cdf(math.log2(x), loc=self.log2_space_mean, scale=self.log2_space_std)
        return (cdf_x - cdf_min) / (cdf_max - cdf_min)

    def min(self):
        """
        Calculate minimum value that can be sampled in the LogNormal distribution.

        :return: minimal value return from distribution.
        """
        return self.hard_clip_min if self.hard_clip_min is not None else 0.

    def max(self):
        """
        Calculate maximal value that can be sampled in the LogNormal distribution.

        :return: maximal value return from distribution.
        """
        return self.hard_clip_max if self.hard_clip_max is not None else np.inf

    def _pseudo_min(self) -> float:
        """
        This is like the function min, but in case it is of 0, there will be a clipping to at least 4 sigmas.
        """
        x = self.min()
        if math.isinf(x) or math.isnan(x):
            x = self.mean() - 4 * self.std()
        elif x == 0:
            x = self._icdf(0.0001, 0.0, self.mean())
        return x

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.

        :return: mean value of the random variable.
        """
        # More info: https://en.wikipedia.org/wiki/Log-normal_distribution#Generation_and_parameters
        # More info: https://en.wikipedia.org/wiki/Log-normal_distribution#Arithmetic_moments
        const = math.exp(math.log(2) * self.log2_space_mean + math.log(2) ** 2 * self.log2_space_std ** 2 / 2)

        if self.hard_clip_min is None and self.hard_clip_max is None:
            mean = const
            return mean

        if self.hard_clip_min == self.hard_clip_max:
            return self.hard_clip_min

        mean_to_calculate_cdf = self.log2_space_mean + math.log(2) * self.log2_space_std ** 2

        cdf_min = 0.
        log_normal_cdf_min = 0
        if self.hard_clip_min is not None:
            log_normal_cdf_min = norm.cdf(
                math.log2(self.hard_clip_min), loc=mean_to_calculate_cdf, scale=self.log2_space_std)
            cdf_min = norm.cdf(
                math.log2(self.hard_clip_min), loc=self.log2_space_mean, scale=self.log2_space_std)

        cdf_max = 1.
        log_normal_cdf_max = 1
        if self.hard_clip_max is not None:
            log_normal_cdf_max = norm.cdf(
                math.log2(self.hard_clip_max), loc=mean_to_calculate_cdf, scale=self.log2_space_std)
            cdf_max = norm.cdf(
                math.log2(self.hard_clip_max), loc=self.log2_space_mean, scale=self.log2_space_std)

        const_norm = const / (cdf_max - cdf_min)
        mean = const_norm * (log_normal_cdf_max - log_normal_cdf_min)
        return mean

    def var(self):
        """
        Calculate variance value (also called esperance) of the random variable.

        :return: variance value of the random variable.
        """
        # Use the mean**2 to calculate variance, with variance = moment_2 - mean**2.
        mean_squared = self.mean() ** 2
        const_moment_2 = math.exp(
            2 * math.log(2) * self.log2_space_mean + 2 * math.log(2) ** 2 * self.log2_space_std ** 2)
        if self.hard_clip_min is None and self.hard_clip_max is None:
            variance = const_moment_2 - mean_squared
            return variance

        if self.hard_clip_min == self.hard_clip_max:
            return 0.

        mean_to_calculate_cdf_for_variance = self.log2_space_mean + 2 * math.log(2) * self.log2_space_std ** 2

        if self.hard_clip_min is None:
            cdf_min = 0.
            log_normal_cdf_min = 0
        else:
            cdf_min = norm.cdf(
                math.log2(self.hard_clip_min), loc=self.log2_space_mean, scale=self.log2_space_std)
            log_normal_cdf_min = norm.cdf(
                math.log2(self.hard_clip_min), loc=mean_to_calculate_cdf_for_variance, scale=self.log2_space_std)

        if self.hard_clip_max is None:
            cdf_max = 1.
            log_normal_cdf_max = 1
        else:
            cdf_max = norm.cdf(
                math.log2(self.hard_clip_max), loc=self.log2_space_mean, scale=self.log2_space_std)
            log_normal_cdf_max = norm.cdf(
                math.log2(self.hard_clip_max), loc=mean_to_calculate_cdf_for_variance, scale=self.log2_space_std)

        const_norm = const_moment_2 / (cdf_max - cdf_min)
        moment_2 = const_norm * (log_normal_cdf_max - log_normal_cdf_min)
        variance = moment_2 - mean_squared
        return variance


class DistributionMixture(DiscreteHyperparameterDistribution):
    """Get a mixture of multiple distribution"""

    def __init__(self, distributions: List[HyperparameterDistribution],
                 distribution_amplitudes: List[float]):
        """
        Create a mixture of multiple distributions.

        Distribution amplitude are normalized to make sure that the sum equals one.
        This normalization ensure to keep a random variable at the end (0 < probability < 1).

        :param distributions: list of multiple instantiated distribution.
        :param distribution_amplitudes: list of float representing the amplitude in the probability distribution function for each distribution.
        """
        DiscreteHyperparameterDistribution.__init__(self, null_default_value=None)

        self.distributions: List[HyperparameterDistribution] = distributions

        # Normalize distribution amplitude
        distribution_amplitudes = np.array(distribution_amplitudes)
        distribution_amplitudes = distribution_amplitudes / np.sum(distribution_amplitudes)
        self.distribution_amplitudes: List[float] = distribution_amplitudes

    @staticmethod
    def build_gaussian_mixture(
            distribution_amplitudes: List[float],
            means: List[float],
            stds: List[float],
            distributions_mins: List[float],
            distributions_max: List[float],
            use_logs: bool = False,
            use_quantized_distributions: bool = False
    ) -> 'DistributionMixture':
        """
        Create a gaussian mixture.

         Will create a mixture distribution which consist of multiple gaussians of different amplitudes, means, standard deviations, mins and max.

        :param distribution_amplitudes: list of different amplitudes to infer to the mixture. The amplitudes are normalized to sum to 1.
        :param means: list of means to infer mean to each gaussian.
        :param stds: list of standard deviations to infer standard deviation to each gaussian.
        :param distributions_mins: list of minimum value that will clip each gaussian. If it is -Inf or None, it will not be clipped.
        :param distributions_max: list of maximal value that will clip each gaussian. If it is +Inf or None, it will not be clipped.
        :param distributions_max: bool weither to use a quantized version or not.
        :param use_logs: use logs boolean
        :param use_quantized_distributions: use quantized distributions boolean

        :return DistributionMixture instance
        """

        distribution_ctor: Callable[[float, float, float, float], HyperparameterDistribution] = \
            LogNormal.from_regular_mean_std if use_logs else Normal

        distributions = []

        for mean, std, single_min, single_max in zip(means, stds, distributions_mins, distributions_max):

            if single_min is None or np.isneginf(single_min):
                single_min = None

            if single_max is None or np.isposinf(single_max):
                single_max = None

            distribution_instance = distribution_ctor(
                mean=mean, std=std, hard_clip_min=single_min, hard_clip_max=single_max)

            if use_quantized_distributions:
                distribution_instance = Quantized(distribution_instance)

            distributions.append(distribution_instance)

        return DistributionMixture(distributions, distribution_amplitudes)

    def rvs(self) -> float:
        """
        Will return a float value drawned from the distribution mixture.

        :return: a float.
        """
        # Draw from which distribution to draw first using the amplitude as probabilities on how to draw from each.
        distribution_index = np.random.choice(len(self.distributions), p=self.distribution_amplitudes, replace=True)

        # Draw from distribution.
        return self.distributions[distribution_index].rvs()

    def pdf(self, x) -> float:
        """
        Calculate the mixture probability distribution value at position `x`.

        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """
        pdf_result = 0

        for distribution_amplitude, distribution in zip(self.distribution_amplitudes, self.distributions):
            pdf_result += (distribution_amplitude * distribution.pdf(x))

        return pdf_result

    def cdf(self, x) -> float:
        """
        Calculate the mixture cumulative distribution value at position `x`.

        :param x: value where the cumulative distribution function is evaluated.
        :return: value of the cumulative distribution function.
        """
        cdf_result = 0

        for distribution_amplitude, distribution in zip(self.distribution_amplitudes, self.distributions):
            cdf_result += (distribution_amplitude * distribution.cdf(x))

        return cdf_result

    def mean(self) -> float:
        """
        Calculate mean of the mixture.

        Mean of the distribution mixture is calculated using the following equation:

        .. math:: mean = \sum_{i=1}^n w_i * \mu_i,

        where :math:`w_i` and :math:`\mu_i` are respectively the amplitude and mean of each distribution.

        :return: mean value of the mixture.
        """
        mean_result = 0

        for distribution_amplitude, distribution in zip(self.distribution_amplitudes, self.distributions):
            mean_result += (distribution_amplitude * distribution.mean())

        return mean_result

    def var(self) -> float:
        """
        Calculate variance of the mixture.

        Variance of a mixture is calculated using the following equation:

        .. math:: variance = (\sum_{i=1}^n w_i * \sigma_i * \mu_i) - \mu^2,

        where :math:`w_i`, :math:`\sigma_i` and :math:`\mu_i` are respectively the amplitude,
        standard deviation and mean of each distribution and :math:`\mu` is the mean of the whole mixture.

        :return: standard deviation value of the mixtture.
        """
        second_moment = 0

        for distribution_amplitude, distribution in zip(self.distribution_amplitudes, self.distributions):
            second_moment += distribution_amplitude * (distribution.var() + distribution.mean() ** 2)

        var_result = second_moment - self.mean() ** 2
        return var_result

    def std(self) -> float:
        """
        Calculate standard deviation of the mixture.

        :return: standard deviation value of the mixtture.
        """
        return math.sqrt(self.var())

    def min(self) -> float:
        """
        Calculate minimal domain value of the mixture.

        :return: minimal domain value of the mixtture.
        """
        return min([distribution.min() for distribution in self.distributions])

    def max(self) -> float:
        """
        Calculate minimal domain value of the mixture.

        :return: minimal domain value of the mixture.
        """
        return max([distribution.max() for distribution in self.distributions])

    def is_discrete(self) -> bool:
        """
        Check if the distribution is discrete.

        :return: True if the distribution is discrete, False otherwise.
        """
        return all([distribution.is_discrete() for distribution in self.distributions])

    def values(self) -> List[float]:
        """
        Get all the values of the distribution.

        :return: list of all the values of the distribution.
        """
        if self.is_discrete():
            return sum([distribution.values() for distribution in self.distributions], [])
        else:
            raise RuntimeError("The distribution is not discrete.")

    def probabilities(self) -> List[float]:
        """
        Get all the probabilities of the distribution.

        :return: list of all the probabilities of the distribution.
        """
        if self.is_discrete():
            return sum([distribution.probabilities() for distribution in self.distributions], [])


class LogDistributionMixture(DistributionMixture):

    @abstractmethod
    def rvs(self) -> float:
        # return super().rvs()
        raise NotImplementedError("TODO: must code this for the 'use logs' method.")

    @staticmethod
    def vals_to_log_wrap(vals: List[float]) -> List[float]:
        return


def _calculate_sum(
        eval_func,
        limits: List[float],
        value_step: float = 1.,
        tol: float = 1e-10,
        number_value_before_stop: int = 5
):
    method, starting_value, stop_value = _get_sum_starting_info(limits)

    current_x = starting_value
    number_negligible_result = 0
    total_sum = 0

    if method.lower() == "increasing":
        comparison = np.greater
    else:
        comparison = np.less

    while True:
        reach_end = comparison(current_x, stop_value)

        if reach_end:
            break

        current_result = eval_func(current_x)
        total_sum += current_result

        is_negligible = abs(current_result) < tol

        if is_negligible:
            number_negligible_result += 1
        else:
            number_negligible_result = 0

        if is_negligible and number_negligible_result > number_value_before_stop:
            break

        if method.lower() == "increasing":
            current_x += value_step
        else:
            current_x -= value_step

    return total_sum


def _get_sum_starting_info(limits):
    if np.isinf(limits[0]) and np.isinf(limits[1]):
        raise ValueError("Cannot calculate a sum on infinite terms.")
    if np.isposinf(limits[0]):
        starting_value = limits[1]
        stop_value = limits[0]
        method = "increasing"
    elif np.isposinf(limits[1]):
        starting_value = limits[0]
        stop_value = limits[1]
        method = "increasing"
    elif np.isneginf(limits[0]):
        starting_value = limits[1]
        stop_value = limits[0]
        method = "decreasing"
    elif np.isneginf(limits[1]):
        starting_value = limits[0]
        stop_value = limits[1]
        method = "decreasing"
    elif np.greater(limits[1], limits[0]):
        starting_value = limits[0]
        stop_value = limits[1]
        method = "increasing"
    else:
        starting_value = limits[1]
        stop_value = limits[0]
        method = "increasing"
    return method, starting_value, stop_value


def get_index_in_list_with_bool(choice_list: List[Any], value: Any) -> int:
    for choice_index, choice in enumerate(choice_list):
        if choice == value and not isinstance(choice, bool) and not isinstance(value, bool):
            index = choice_index
            return index

        if choice is value:
            index = choice_index
            return index

    raise ValueError("{} is not in list".format(value))



import numpy as np
import pytest
from pytest import skip

from neuraxle.base import CX, ExecutionPhase
from neuraxle.data_container import DataContainer as DACT
from neuraxle.steps.flow import IfExecutionPhaseIsThen, ExecutionPhaseSwitch
from testing.test_forcehandle_mixin import ForceHandleIdentity


class SomeStep(ForceHandleIdentity):
    def __init__(self):
        ForceHandleIdentity.__init__(self)
        self.did_process = False

    def _did_process(self, data_container: DACT, context: CX) -> DACT:
        self.did_process = True
        return data_container


def test_ifexecphase_same_then_execute_step(tmpdir):
    _run(tmpdir, ExecutionPhase.TRAIN, True)


def test_ifexecphase_different_then_skip_step(tmpdir):
    _run(tmpdir, ExecutionPhase.TEST, False)


def _run(tmpdir, phase, expected):
    context = CX(root=tmpdir, execution_phase=phase)
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    some_step = SomeStep()
    p = IfExecutionPhaseIsThen(ExecutionPhase.TRAIN, some_step)
    p = p.with_context(context)

    p.fit_transform(data_inputs)
    assert some_step.did_process is expected


def test_ifexecphase_raise_exception_when_unspecified(tmpdir):
    context = CX(root=tmpdir)
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    some_step = SomeStep()
    p = IfExecutionPhaseIsThen(ExecutionPhase.TRAIN, some_step)
    p = p.with_context(context)

    with pytest.raises(ValueError) as error_info:
        p.fit_transform(data_inputs)
    assert some_step.did_process is False


def test_execswitch(tmpdir):
    context = CX(root=tmpdir, execution_phase=ExecutionPhase.TRAIN)
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    phase_to_step = {p: SomeStep() for p in (ExecutionPhase.PRETRAIN, ExecutionPhase.TRAIN, ExecutionPhase.TEST)}
    p = ExecutionPhaseSwitch(phase_to_step)
    p_c = p.with_context(context)

    p_c.fit_transform(data_inputs)
    assert phase_to_step[ExecutionPhase.PRETRAIN].did_process is False
    assert phase_to_step[ExecutionPhase.TRAIN].did_process is True
    assert phase_to_step[ExecutionPhase.TEST].did_process is False

    p_c = p.with_context(context.set_execution_phase(ExecutionPhase.UNSPECIFIED))
    with pytest.raises(KeyError) as error_info:
        p_c.fit_transform(data_inputs)



"""
Data Steps
====================================
You can find here steps that take action on data.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import random
from operator import attrgetter
from typing import Iterable, List, Tuple

from neuraxle.base import BaseStep, BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import ForceHandleOnlyMixin, MetaStep
from neuraxle.data_container import DACT, _inner_concatenate_np_array
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import TrainOnlyWrapper
from neuraxle.steps.output_handlers import InputAndOutputTransformerMixin


class DataShuffler(InputAndOutputTransformerMixin, BaseTransformer):
    """
    Data Shuffling step that shuffles data inputs, and expected_outputs at the same time.

    .. code-block:: python

        p = Pipeline([
            TrainOnlyWrapper(DataShuffler(seed=42, increment_seed_after_each_fit=True, increment_seed_after_each_fit=False)),
            EpochRepeater(ForecastingPipeline(), epochs=EPOCHS, repeat_in_test_mode=False)
        ])

    .. warning::
        You probably always want to wrap this step by a :class:`TrainOnlyWrapper`

    .. seealso::
        :class:`EpochRepeater`,
        :class:`~neuraxle.steps.flow.TrainOnlyWrapper`,
        :class:`~neuraxle.steps.output_handlers.InputAndOutputTransformerMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, seed=None, increment_seed_after_each_fit=True):
        BaseTransformer.__init__(self)
        InputAndOutputTransformerMixin.__init__(self)
        if seed is None:
            seed = 42
        self.seed = seed
        self.increment_seed_after_each_fit = increment_seed_after_each_fit

    def transform(self, data_inputs):
        """
        Shuffle data inputs, and expected outputs.

        :param data_inputs: (data inputs, expected outputs) tuple to shuffle
        :return:
        """
        if self.increment_seed_after_each_fit:
            self.seed += 1

        di, eo = data_inputs
        data = list(zip(di, eo))
        random.Random(self.seed).shuffle(data)

        data_inputs_shuffled, expected_outputs_shuffled = list(zip(*data))

        return list(data_inputs_shuffled), list(expected_outputs_shuffled)


class EpochRepeater(ForceHandleOnlyMixin, MetaStep):
    """
    Repeat wrapped step fit, or transform for the number of epochs passed in the constructor.

    .. code-block:: python

        p = Pipeline([
            TrainOnlyWrapper(DataShuffler(seed=42, increment_seed_after_each_fit=True, increment_seed_after_each_fit=False)),
            EpochRepeater(ForecastingPipeline(), epochs=EPOCHS, repeat_in_test_mode=False)
        ])

    .. seealso::
        :class:`DataShuffler`,
        :class:`~neuraxle.base.MetaStepMixin`,
        :class:`~neuraxle.steps.flow.TrainOnlyWrapper`,
        :class:`~neuraxle.steps.flow.TestOnlyWrapper`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, wrapped, epochs, repeat_in_test_mode=False, cache_folder_when_no_handle=None):
        BaseStep.__init__(self)
        MetaStep.__init__(self, wrapped)
        ForceHandleOnlyMixin.__init__(self, cache_folder=cache_folder_when_no_handle)

        self.repeat_in_test_mode = repeat_in_test_mode
        self.epochs = epochs

    def _fit_transform_data_container(self, data_container: DACT, context: CX) -> Tuple['BaseStep', DACT]:
        """
        Fit transform wrapped step self.epochs times using wrapped step handle fit transform method.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: (fitted self, data container)
        """
        if self._should_repeat():
            for _ in range(self.epochs - 1):
                self.wrapped = self.wrapped.handle_fit(data_container.copy(), context)

        self.wrapped, data_container = self.wrapped.handle_fit_transform(data_container, context)
        return self, data_container

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Iterable]:
        """
        Fit transform wrapped step self.epochs times.

        :param data_inputs: data inputs to fit on
        :param expected_outputs: expected_outputs to fit on
        :return: fitted self
        """
        if self._should_repeat():
            for _ in range(self.epochs - 1):
                self.wrapped = self.wrapped.fit(data_inputs, expected_outputs)

        self.wrapped, outputs = self.wrapped.fit_transform(data_inputs, expected_outputs)

        return self, outputs

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'BaseStep':
        """
        Fit wrapped step self.epochs times using wrapped step handle fit method.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: (fitted self, data container)
        """

        if self._should_repeat():
            for _ in range(self.epochs):
                self.wrapped = self.wrapped.handle_fit(data_container.copy(), context)
        return self

    def fit(self, data_inputs, expected_outputs=None) -> 'BaseStep':
        """
        Fit wrapped step self.epochs times.

        :param data_inputs: data inputs to fit on
        :param expected_outputs: expected_outputs to fit on
        :return: fitted self
        """
        if self._should_repeat():
            for _ in range(self.epochs):
                self.wrapped = self.wrapped.fit(data_inputs, expected_outputs)
        else:
            self.wrapped = self.wrapped.fit(data_inputs, expected_outputs)

        return self

    def _should_repeat(self):
        return self.is_train or (not self.is_train and self.repeat_in_test_mode)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        return self.wrapped.handle_transform(data_container, context)

    def _get_epochs(self):
        epochs = self.epochs
        if not self._should_repeat_fit():
            epochs = 1
        return epochs

    def _should_repeat_fit(self):
        return self.is_train or (not self.is_train and self.repeat_in_test_mode)


class TrainShuffled(Pipeline):
    def __init__(self, wrapped, seed=None):
        Pipeline.__init__(self, [
            TrainOnlyWrapper(DataShuffler(seed=seed)),
            wrapped
        ])


class InnerConcatenateDataContainer(ForceHandleOnlyMixin, BaseTransformer):
    """
    Concatenate inner features of sub data containers along `axis=-1`..

    Code example:

    .. code-block:: python

        data_container = DataContainer(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)
        data_container.add_sub_data_container(name='1d_data_source', data_container=data_container_1d)
        data_container.add_sub_data_container(name='2d_data_source', data_container=data_container_2d)

        # data container with sub data containers :
        # DataContainer(data_inputs=data_inputs_3d, expected_outputs=expected_outputs, sub_data_containers=[('1d_data_source', data_container_1d), ('2d_data_source', data_container_2d)])

        p = Pipeline([
            InnerConcatenateDataContainer()
            # is equivalent to ZipData(sub_data_container_names=['1d_data_source', '2d_data_source'])
        ])

        data_container = p.handle_transform(data_container, ExecutionContext())

        # new_shape: (batch_size, time_steps, n_features + batch_features + 1)


    .. seealso::
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.base.BaseStep`
        :class:`~neuraxle.data_container.DataContainer`
    """

    def __init__(self, sub_data_container_names=None):
        BaseTransformer.__init__(self)
        ForceHandleOnlyMixin.__init__(self)

        self.data_sources = sub_data_container_names

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseTransformer', DACT]:
        """
        Merge sub data containers into the current data container.

        :param data_container: data container to zip
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: base step, data container
        :rtype: Tuple[BaseTransformer, DataContainer]
        """
        return self, self._concatenate_sub_data_containers(data_container)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Merge sub data containers into the current data container.

        :param data_container: data container to zip
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: base step, data container
        :rtype: DataContainer
        """
        return self._concatenate_sub_data_containers(data_container)

    def _concatenate_sub_data_containers(self, data_container: DACT) -> DACT:
        """
        Merge sub data containers into the current data container.

        :param data_container: data container to zip
        :type data_container: DataContainer
        :return: base step, data container
        :rtype: DataContainer
        """
        sub_data_containers_to_zip = [data_container]
        if self.data_sources is None:
            self.data_sources = data_container.get_sub_data_container_names()

        for name, sub_data_container in data_container.sub_data_containers:
            if name in self.data_sources:
                sub_data_containers_to_zip.append(sub_data_container)

        if len(sub_data_containers_to_zip) > 1:
            data_container = self._concatenate_sub_data_container(sub_data_containers_to_zip)

        return data_container

    def _concatenate_sub_data_container(self, data_container_to_zip: List[DACT]) -> DACT:
        """
        Zip a data container into another data container with a higher dimension.

        :param data_container: data container
        :type data_container: DataContainer
        :param data_container_to_zip: data container to concatenate
        :type data_container_to_zip: DataContainer
        :return: concatenated data containers
        """
        data_inputs = _inner_concatenate_np_array(list(map(attrgetter("data_inputs"), data_container_to_zip)))

        expected_outputs = _inner_concatenate_np_array(list(map(attrgetter("expected_outputs"), data_container_to_zip)))

        return data_container_to_zip[0].copy()\
            .set_data_inputs(data_inputs)\
            .set_expected_outputs(expected_outputs)


class ZipBatchDataContainer(ForceHandleOnlyMixin, BaseTransformer):
    """
    WARNING: Unexpected behaviour from this class. It's not to date.

    Concatenate outer batch of sub data containers along `axis=0`..

    Code example:

    .. code-block:: python

        data_container = DataContainer(data_inputs=data_inputs_3d, expected_outputs=expected_outputs_3d)
        data_container.add_sub_data_container(name='1d_data_source', data_container=data_container_1d)
        data_container.add_sub_data_container(name='2d_data_source', data_container=data_container_2d)

        # data container with sub data containers :
        # DataContainer(data_inputs=data_inputs_3d, expected_outputs=expected_outputs, sub_data_containers=[('1d_data_source', data_container_1d), ('2d_data_source', data_container_2d)])

        p = Pipeline([
            ZipBatchDataContainer()
            # is equivalent to ZipBatchDataContainer(sub_data_container_names=['2d_data_source'])
        ])

        data_container = p.handle_transform(data_container, ExecutionContext())

        # new_shape: (batch_size, ((time_steps, n_features_3d), n_features_2d))


    .. seealso::
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.base.BaseStep`
        :class:`~neuraxle.data_container.DataContainer`
    """

    def __init__(self, sub_data_container_names=None):
        BaseTransformer.__init__(self)
        ForceHandleOnlyMixin.__init__(self)

        self.data_sources = sub_data_container_names

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Merge sub data containers into the current data container.

        :param data_container: data container to zip
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: base step, data container
        :rtype: DataContainer
        """
        return self._batch_zip_sub_data_containers(data_container)

    def _batch_zip_sub_data_containers(self, data_container: DACT):
        """
        Zip sub data containers on the batch dimension.

        :param data_container: data container to zip
        :type data_container: DataContainer
        :return: base step, data container
        :rtype: DataContainer
        """
        sub_data_containers_to_zip = []
        if self.data_sources is None:
            self.data_sources = data_container.get_sub_data_container_names()

        for name, sub_data_container in data_container.sub_data_containers:
            if name in self.data_sources:
                sub_data_containers_to_zip.append(sub_data_container)

        for data_container_to_zip in sub_data_containers_to_zip:
            data_container = self._batch_zip_sub_data_container(data_container, data_container_to_zip)

        return data_container

    def _batch_zip_sub_data_container(self, data_container, data_container_to_zip) -> DACT:
        """
        Zip sub data container on the batch dimension.

        :param data_container: data container
        :type data_container: DataContainer
        :param data_container_to_zip: data container to concatenate
        :type data_container_to_zip: DataContainer
        :return: concatenated data containers
        """
        new_data_inputs = []
        for di, other_di in zip(data_container.data_inputs, data_container_to_zip.data_inputs):
            new_data_inputs.append((di, other_di))

        new_expected_outputs = []
        for eo, other_eo in zip(data_container.expected_outputs, data_container_to_zip.expected_outputs):
            new_expected_outputs.append((eo, other_eo))

        data_container.set_data_inputs(new_data_inputs)
        data_container.set_expected_outputs(new_expected_outputs)

        return data_container




from neuraxle.base import _RecursiveArguments
from neuraxle.hyperparams.space import HyperparameterSamples


def test_recursive_arguments_should_get_root_level():
    ra = _RecursiveArguments(kwargs={'hyperparams': HyperparameterSamples({
        'hp0': 0,
        'hp1': 1,
        'pipeline__stepa__hp2': 2,
        'pipeline__stepb__hp3': 3
    })})

    root_ra = ra[None]

    root_ra.args == []
    root_ra.kwargs == {'hyperparams': HyperparameterSamples({
        'hp0': 0,
        'hp1': 1
    })}


def test_recursive_arguments_should_get_recursive_levels():
    ra = _RecursiveArguments(kwargs={'hyperparams': HyperparameterSamples({
        'hp0': 0,
        'hp1': 1,
        'stepa__hp2': 2,
        'stepb__hp3': 3,
        'stepb__stepd__hp4': 4
    })})

    ra = ra['stepb']

    ra.args == []
    ra.kwargs == {'hyperparams': HyperparameterSamples({
        'stepb__hp3': 2,
        'stepb__stepd__hp4': 4
    })}


def test_recursive_arguments_should_have_copy_constructor():
    ra = _RecursiveArguments(
        ra=_RecursiveArguments(kwargs={'hyperparams': HyperparameterSamples({
            'hp0': 0,
            'hp1': 1
        })}),
    )

    ra.args == []
    ra.kwargs == {'hyperparams': HyperparameterSamples({
        'hp0': 0,
        'hp1': 1,
    })}



"""
Boston Housing Regression
==========================

This example solves a regression problem using a pipeline with the following steps:

 - Feature augmentation with PCA and Fast ICA,
 - A Pre-regression using an ensemble containing gradient boosted, and a KMeans clustering for even more features in the stacking,
 - The model stacking using a ridge regression.

This example also prints the shapes of the objects between the pipeline elements.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np
from sklearn.cluster import KMeans
from sklearn.datasets import load_boston
from sklearn.decomposition import PCA, FastICA
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import NumpyShapePrinter
from neuraxle.steps.sklearn import RidgeModelStacking
from neuraxle.union import AddFeatures


def main():
    boston = load_boston()
    X, y = shuffle(boston.data, boston.target, random_state=13)
    X = X.astype(np.float32)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, shuffle=False)

    p = Pipeline([
        NumpyShapePrinter(),
        AddFeatures([
            PCA(n_components=2),
            FastICA(n_components=2),
        ]),
        NumpyShapePrinter(),
        RidgeModelStacking([
            GradientBoostingRegressor(),
            GradientBoostingRegressor(n_estimators=500),
            GradientBoostingRegressor(max_depth=5),
            KMeans(),
        ]),
        NumpyShapePrinter(),
    ])

    print("Fitting on train:")
    p = p.fit(X_train, y_train)
    print("")
    print("Transforming train and test:")
    y_train_predicted = p.predict(X_train)
    y_test_predicted = p.predict(X_test)
    print("")
    print("Evaluating transformed train:")
    score_train = r2_score(y_train_predicted, y_train)
    print('R2 regression score:', score_train)
    print("")
    print("Evaluating transformed test:")
    score_test = r2_score(y_test_predicted, y_test)
    print('R2 regression score:', score_test)

    assert y_train_predicted.shape == (379,)
    assert y_test_predicted.shape == (127,)
    assert isinstance(score_train, float)
    assert isinstance(score_test, float)

    return y_train_predicted, y_test_predicted, score_train, score_test


if __name__ == "__main__":
    main()



"""
Tests for Hyperparameters Distributions
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import random
from collections import Counter

import numpy as np
import pytest
from neuraxle.hyperparams.distributions import *


NUM_TRIALS = 5000


def hd_rvs_many(hd: HyperparameterDistribution):
    random.seed(111)
    np.random.seed(111)
    return hd.rvs_many(NUM_TRIALS)


def test_boolean_distribution():
    hd = Boolean()

    samples = hd_rvs_many(hd)
    falses = Counter(samples).get(False)
    trues = Counter(samples).get(True)

    # You'd need to win the lotto for this test to fail. Or a broken random sampler. Or a bug.
    assert trues > NUM_TRIALS * 0.4
    assert falses > NUM_TRIALS * 0.4
    assert hd.pdf(False) == 0.5
    assert hd.pdf(0.) == 0.5
    assert hd.pdf(True) == 0.5
    assert hd.pdf(1.) == 0.5
    assert hd.pdf(-0.1) == 0.
    assert hd.pdf(1.1) == 0.

    assert hd.cdf(False) == 0.5
    assert hd.cdf(0.) == 0.5
    assert hd.cdf(True) == 1.
    assert hd.cdf(1.) == 1.
    assert hd.cdf(-0.1) == 0.
    assert hd.cdf(1.1) == 1.

    assert hd.min() == 0
    assert hd.max() == 1
    assert abs(hd.mean() - 0.5) < 1e-6
    assert abs(hd.std() - 0.5) < 1e-6
    assert abs(hd.var() - 0.25) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-2
    assert abs(hd.var() - np.var(samples)) < 1e-2


def test_boolean_distribution_with_proba():
    proba_is_true = 0.7
    hd = Boolean(proba_is_true=proba_is_true)

    samples = hd_rvs_many(hd)
    falses = Counter(samples).get(False)
    trues = Counter(samples).get(True)

    # You'd need to win the lotto for this test to fail. Or a broken random sampler. Or a bug.
    assert trues > NUM_TRIALS * (proba_is_true - 0.1)
    assert falses > NUM_TRIALS * (1 - proba_is_true - 0.1)
    assert abs(hd.pdf(False) - (1 - proba_is_true)) < 1e-6
    assert abs(hd.pdf(0.) - (1 - proba_is_true)) < 1e-6
    assert abs(hd.pdf(True) - proba_is_true) < 1e-6
    assert abs(hd.pdf(1.) - proba_is_true) < 1e-6
    assert abs(hd.pdf(-0.1) - 0.) < 1e-6
    assert abs(hd.pdf(1.1) - 0.) < 1e-6

    assert abs(hd.cdf(False) - (1 - proba_is_true)) < 1e-6
    assert abs(hd.cdf(0.) - (1 - proba_is_true)) < 1e-6
    assert abs(hd.cdf(True) - 1.) < 1e-6
    assert abs(hd.cdf(1.) - 1.) < 1e-6
    assert abs(hd.cdf(-0.1) - 0.) < 1e-6
    assert abs(hd.cdf(1.1) - 1.) < 1e-6

    assert hd.min() == 0
    assert hd.max() == 1
    assert abs(hd.mean() - proba_is_true) < 1e-6
    assert abs(hd.std() - math.sqrt(proba_is_true * (1 - proba_is_true))) < 1e-6
    assert abs(hd.var() - proba_is_true * (1 - proba_is_true)) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-2
    assert abs(hd.var() - np.var(samples)) < 1e-2


@pytest.mark.parametrize("ctor", [Choice, PriorityChoice])
def test_choice_and_priority_choice(ctor):
    choice_list = [0, 1, False, "Test"]
    hd = ctor(choice_list)

    samples = hd_rvs_many(hd)
    z0 = Counter(samples).get(0)
    z1 = Counter(samples).get(1)
    zNone = Counter(samples).get(False)
    zTest = Counter(samples).get("Test")

    # You'd need to win the lotto for this test to fail. Or a broken random sampler. Or a bug.
    assert z0 > NUM_TRIALS * 0.2
    assert z1 > NUM_TRIALS * 0.2
    assert zNone > NUM_TRIALS * 0.2
    assert zTest > NUM_TRIALS * 0.2

    assert (hd.pdf(0) - 1 / 4) < 1e-6
    assert (hd.pdf(1) - 1 / 4) < 1e-6
    assert (hd.pdf(False) - 1 / 4) < 1e-6
    assert (hd.pdf("Test") - 1 / 4) < 1e-6
    assert abs(hd.pdf(0) - 1 / 4) < 1e-6
    assert abs(hd.pdf(1) - 1 / 4) < 1e-6
    assert abs(hd.pdf(False) - 1 / 4) < 1e-6
    assert abs(hd.pdf("Test") - 1 / 4) < 1e-6

    assert abs(hd.cdf(0) - 1 / 4) < 1e-6
    assert abs(hd.cdf(1) - 2 / 4) < 1e-6
    assert abs(hd.cdf(False) - 3 / 4) < 1e-6
    assert hd.cdf("Test") == 1.

    with pytest.raises(ValueError):
        assert hd.pdf(3) == 0.
        assert hd.cdf(3) == 0.

    assert hd.min() == 0
    assert hd.max() == len(choice_list) - 1
    assert abs(hd.mean() - (len(choice_list) - 1) / 2) < 1e-6
    assert abs(hd.var() - (len(choice_list) ** 2 - 1) / 12) < 1e-6
    assert abs(hd.std() - math.sqrt((len(choice_list) ** 2 - 1) / 12)) < 1e-6
    # Convert samples in sample index
    samples_index = [get_index_in_list_with_bool(choice_list, sample) for sample in samples]
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs((hd.mean() - np.mean(samples_index)) / hd.mean()) < 1e-1
    assert abs((hd.var() - np.var(samples_index)) / hd.var()) < 1e-1


@pytest.mark.parametrize("ctor", [Choice, PriorityChoice])
def test_choice_and_priority_choice_with_probas(ctor):
    probas = [0.1, 0.4, 0.3, 0.2]
    probas_array = np.array(probas)
    choice_list = [0, 1, False, "Test"]
    hd = ctor(choice_list, probas=probas)

    samples = hd_rvs_many(hd)
    z0 = Counter(samples).get(0)
    z1 = Counter(samples).get(1)
    zNone = Counter(samples).get(False)
    zTest = Counter(samples).get("Test")

    # You'd need to win the lotto for this test to fail. Or a broken random sampler. Or a bug.
    assert z0 > NUM_TRIALS * (probas[0] - 0.05)
    assert z1 > NUM_TRIALS * (probas[1] - 0.05)
    assert zNone > NUM_TRIALS * (probas[2] - 0.05)
    assert zTest > NUM_TRIALS * (probas[3] - 0.05)

    assert abs(hd.pdf(0) - probas[0]) < 1e-6
    assert abs(hd.pdf(1) - probas[1]) < 1e-6
    assert abs(hd.pdf(False) - probas[2]) < 1e-6
    assert abs(hd.pdf("Test") - probas[3]) < 1e-6

    assert abs(hd.cdf(0) - probas_array[0]) < 1e-6
    assert abs(hd.cdf(1) - np.sum(probas_array[0:2])) < 1e-6
    assert abs(hd.cdf(False) - np.sum(probas_array[0:3])) < 1e-6
    assert abs(hd.cdf("Test") - 1.) < 1e-6

    with pytest.raises(ValueError):
        assert hd.pdf(3) == 0.
        assert hd.cdf(3) == 0.

    assert hd.min() == 0
    assert hd.max() == len(choice_list) - 1
    assert abs(hd.mean() - 1.6) < 1e-6
    assert abs(hd.var() - 0.84) < 1e-6
    assert abs(hd.std() - 0.9165151389911679) < 1e-6
    # Convert samples in sample index
    samples_index = [get_index_in_list_with_bool(choice_list, sample) for sample in samples]
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs((hd.mean() - np.mean(samples_index)) / hd.mean()) < 1e-1
    assert abs((hd.var() - np.var(samples_index)) / hd.var()) < 1e-1


def test_quantized_uniform():
    low = -10
    high = 10
    hd = Quantized(Uniform(low, high))

    samples = hd_rvs_many(hd)

    for s in samples:
        assert type(s) == int
    samples_mean = np.abs(np.mean(samples))
    assert samples_mean < 1.0
    assert min(samples) >= -10.0
    assert max(samples) <= 10.0

    assert abs(hd.pdf(-10) - 1 / 40) < 1e-6
    assert abs(hd.pdf(-9) - 1 / 20) < 1e-6
    assert abs(hd.pdf(0) - 1 / 20) < 1e-6
    assert abs(hd.pdf(9) - 1 / 20) < 1e-6
    assert abs(hd.pdf(10) - 1 / 40) < 1e-6

    assert abs(hd.cdf(-10) - 1 / 40) < 1e-6
    assert abs(hd.cdf(-9) - 1.5 / 20) < 1e-6
    assert abs(hd.cdf(0) - 10.5 / 20) < 1e-6
    assert abs(hd.cdf(9) - 19.5 / 20) < 1e-6
    assert abs(hd.cdf(9.2) - 19.5 / 20) < 1e-6
    assert hd.cdf(10) == 1.

    assert hd.min() == low
    assert hd.max() == high
    assert abs(hd.mean() - 0.0) < 1e-6
    assert abs(hd.var() - 33.50000000000001) < 1e-6
    assert abs(hd.std() - 5.787918451395114) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 1e-1


def test_randint():
    low = -10
    high = 10
    hd = RandInt(low, high)

    samples = hd_rvs_many(hd)

    for s in samples:
        assert type(s) == int
    samples_mean = np.abs(np.mean(samples))
    assert samples_mean < 1.0
    assert min(samples) >= -10.0
    assert max(samples) <= 10.0

    assert hd.pdf(-11) == 0.
    assert abs(hd.pdf(-10) - 1 / (10 + 10 + 1)) < 1e-6
    assert abs(hd.pdf(0) - 1 / (10 + 10 + 1)) < 1e-6
    assert hd.pdf(0.5) == 0.
    assert abs(hd.pdf(10) - 1 / (10 + 10 + 1)) < 1e-6
    assert hd.pdf(11) == 0.

    assert hd.cdf(-10.1) == 0.
    assert abs(hd.cdf(-10) - 1 / (10 + 10 + 1)) < 1e-6
    assert abs(hd.cdf(0) - (0 + 10 + 1) / (10 + 10 + 1)) < 1e-6
    assert abs(hd.cdf(5) - (5 + 10 + 1) / (10 + 10 + 1)) < 1e-6
    assert abs(hd.cdf(10) - 1.) < 1e-6
    assert hd.cdf(10.1) == 1.

    assert hd.min() == low
    assert hd.max() == high
    assert abs(hd.mean() - (10 - 10) / 2) < 1e-6
    assert abs(hd.var() - ((high - low + 1) ** 2 - 1) / 12) < 1e-6
    assert abs(hd.std() - math.sqrt(((high - low + 1) ** 2 - 1) / 12)) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 1e-1


def test_uniform():
    low = -10
    high = 10
    hd = Uniform(low, high)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert samples_mean < 1.0
    assert min(samples) >= -10.0
    assert max(samples) <= 10.0
    assert hd.pdf(-10.1) == 0.
    assert abs(hd.pdf(0) - 1 / (10 + 10)) < 1e-6
    assert hd.pdf(10.1) == 0.
    assert hd.cdf(-10.1) == 0.
    assert abs(hd.cdf(0) - (0 + 10) / (10 + 10)) < 1e-6
    assert hd.cdf(10.1) == 1.

    assert hd.min() == low
    assert hd.max() == high
    assert abs(hd.mean() - (10 - 10) / 2) < 1e-6
    assert abs(hd.var() - 1 / 12 * (high - low) ** 2) < 1e-6
    assert abs(hd.std() - math.sqrt(1 / 12 * (high - low) ** 2)) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 1e-1


def test_loguniform():
    min_included = 0.001
    max_included = 10
    hd = LogUniform(min_included, max_included)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert samples_mean < 1.15  # if it was just uniform, this assert would break.
    assert min(samples) >= 0.001
    assert max(samples) <= 10.0
    assert hd.pdf(0.0001) == 0.
    assert abs(hd.pdf(2) - 0.054286810237906484) < 1e-6
    assert hd.pdf(10.1) == 0.
    assert hd.cdf(0.0001) == 0.
    assert abs(hd.cdf(2) - (math.log2(2) - math.log2(0.001)) / (math.log2(10) - math.log2(0.001))) < 1e-6
    assert hd.cdf(10.1) == 1.

    assert hd.min() == min_included
    assert hd.max() == max_included
    assert abs(hd.mean() - (max_included - min_included) / (
        math.log(2) * (math.log2(max_included) - math.log2(min_included)))) < 1e-6
    esperance_squared = (max_included ** 2 - min_included ** 2) / (
        2 * math.log(2) * (math.log2(max_included) - math.log2(min_included)))
    assert abs(hd.var() - (esperance_squared - hd.mean() ** 2)) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 5e-2
    assert abs(hd.var() - np.var(samples)) < 2.5e-1


def test_normal():
    hd_mean = 0.0
    hd_std = 1.0
    hd = Normal(hd_mean, hd_std)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert samples_mean < 0.1
    samples_std = np.std(samples)
    assert 0.9 < samples_std < 1.1
    assert abs(hd.pdf(-1.) - 0.24197072451914337) < 1e-6
    assert abs(hd.pdf(0.) - 0.3989422804014327) < 1e-6
    assert abs(hd.pdf(1.) - 0.24197072451914337) < 1e-6
    assert abs(hd.cdf(-1.) - 0.15865525393145707) < 1e-6
    assert abs(hd.cdf(0.) - 0.5) < 1e-6
    assert abs(hd.cdf(1.) - 0.8413447460685429) < 1e-6

    assert hd.min() == -np.inf
    assert hd.max() == np.inf
    assert abs(hd.mean() - hd_mean) < 1e-6
    assert abs(hd.var() - hd_std ** 2) < 1e-6
    assert abs(hd.std() - hd_std) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 5e-2
    assert abs(hd.var() - np.var(samples)) < 5e-2


def test_normal_truncated():
    hd_mean = 2.0
    hd_std = 1.0
    hard_clip_min = 1.8
    hard_clip_max = 2.5
    hd = Normal(hd_mean, hd_std, hard_clip_min=hard_clip_min, hard_clip_max=hard_clip_max)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert 2.0 < samples_mean < 2.2
    samples_std = np.std(samples)
    assert 0 < samples_std < 0.4
    assert abs(hd.pdf(1.7) - 0.) < 1e-6
    assert abs(hd.pdf(1.8) - 1.4444428136247596) < 1e-6
    assert abs(hd.pdf(2.) - 1.473622494051997) < 1e-6
    assert abs(hd.pdf(2.25) - 1.4282838963071145) < 1e-6
    assert abs(hd.pdf(2.5) - 1.3004672865798739) < 1e-6
    assert abs(hd.pdf(2.6) - 0.) < 1e-6
    assert abs(hd.cdf(1.7) - 0.) < 1e-6
    assert abs(hd.cdf(1.8) - 0.) < 1e-6
    assert abs(hd.cdf(2.) - 0.2927714018778846) < 1e-6
    assert abs(hd.cdf(2.25) - 0.65737517785574) < 1e-6
    assert abs(hd.cdf(2.5) - 1.) < 1e-6
    assert abs(hd.cdf(2.6) - 1.) < 1e-6

    assert np.all((np.array(samples) >= hard_clip_min) & (np.array(samples) <= hard_clip_max))
    assert hd.min() == hard_clip_min
    assert hd.max() == hard_clip_max
    assert abs(hd.mean() - 2.1439755270448857) < 1e-6
    assert abs(hd.var() - 0.04014884159725845) < 1e-6
    assert abs(hd.std() - 0.20037175848222336) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-2
    assert abs(hd.var() - np.var(samples)) < 1e-2


def test_normal_onside_lower_tail_truncated():
    hd_mean = 2.0
    hd_std = 1.0
    hard_clip_min = 1.8
    hard_clip_max = None
    hd = Normal(hd_mean, hd_std, hard_clip_min=hard_clip_min, hard_clip_max=hard_clip_max)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert 2.5 < samples_mean < 2.8
    samples_std = np.std(samples)
    assert 0.5 < samples_std < 0.7
    assert abs(hd.pdf(1.7) - 0.) < 1e-6
    assert abs(hd.pdf(1.8) - 0.6750731797902921) < 1e-6
    assert abs(hd.pdf(2.) - 0.688710562638179) < 1e-6
    assert abs(hd.pdf(2.5) - 0.607784938305487) < 1e-6
    assert abs(hd.pdf(5.) - 0.007650883256198442) < 1e-6
    assert abs(hd.cdf(1.7) - 0.) < 1e-6
    assert abs(hd.cdf(1.8) - 0.) < 1e-6
    assert abs(hd.cdf(2.) - 0.13682931532705794) < 1e-6
    assert abs(hd.cdf(2.5) - 0.46735888290117106) < 1e-6
    assert abs(hd.cdf(5.) - 0.9976696151835984) < 1e-6

    assert np.all(np.array(samples) >= hard_clip_min)
    assert hd.min() == hard_clip_min
    assert hd.max() == np.inf
    assert abs(hd.mean() - 2.6750731797902922) < 1e-6
    assert abs(hd.var() - 0.4092615659697655) < 1e-6
    assert abs(hd.std() - 0.6397355437755241) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-2
    assert abs(hd.var() - np.var(samples)) < 1e-2


def test_normal_onside_upper_tail_truncated():
    hd_mean = 2.0
    hd_std = 1.0
    hard_clip_min = None
    hard_clip_max = 2.5
    hd = Normal(hd_mean, hd_std, hard_clip_min=hard_clip_min, hard_clip_max=hard_clip_max)

    samples = hd_rvs_many(hd)

    samples_mean = np.abs(np.mean(samples))
    assert 1.4 < samples_mean < 1.6
    samples_std = np.std(samples)
    assert 0.6 < samples_std < 0.8
    assert abs(hd.pdf(-1.) - 0.006409383965359982) < 1e-6
    assert abs(hd.pdf(1.8) - 0.5655298962361072) < 1e-6
    assert abs(hd.pdf(2.) - 0.5769543579652687) < 1e-6
    assert abs(hd.pdf(2.5) - 0.5091604338370336) < 1e-6
    assert abs(hd.pdf(2.6) - 0.) < 1e-6
    assert abs(hd.cdf(-1.) - 0.0019522361765567414) < 1e-6
    assert abs(hd.cdf(1.8) - 0.6084788605670465) < 1e-6
    assert abs(hd.cdf(2.) - 0.723105053423659) < 1e-6
    assert abs(hd.cdf(2.5) - 1.) < 1e-6
    assert abs(hd.cdf(2.6) - 1.) < 1e-6

    assert np.all(np.array(samples) <= hard_clip_max)
    assert hd.min() == -np.inf
    assert hd.max() == hard_clip_max
    assert abs(hd.mean() - 1.4908395661629665) < 1e-6
    assert abs(hd.var() - 0.486175435696367) < 1e-6
    assert abs(hd.std() - 0.6972628168032245) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-2
    assert abs(hd.var() - np.var(samples)) < 1e-2


@pytest.mark.parametrize("seed", (15, 20, 32, 40, 50))
def test_lognormal(seed):
    np.random.seed(seed)

    log2_space_mean = 0.0
    log2_space_std = 2.0
    hd = LogNormal(log2_space_mean, log2_space_std)

    samples = hd_rvs_many(hd)

    samples_median = np.median(samples)
    assert 0.9 < samples_median < 1.1
    samples_std = np.std(samples)
    assert 5 < samples_std < 8
    assert hd.pdf(0.) == 0.
    assert abs(hd.pdf(1.) - 0.28777602476804065) < 1e-6
    assert abs(hd.pdf(5.) - 0.029336304593386688) < 1e-6
    assert hd.cdf(0.) == 0.
    assert hd.cdf(1.) == 0.5
    assert abs(hd.cdf(5.) - 0.8771717397015799) < 1e-6

    assert hd.min() == 0
    assert hd.max() == np.inf
    assert abs(hd.mean() - 2.614063815405198) < 1e-6
    assert abs(hd.var() - 39.86106421503915) < 1e-6
    assert abs(hd.std() - 6.313561927710787) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 2.5e-1


def test_lognormal_clipped():
    log2_space_mean = 10.0
    log2_space_std = 5.0
    hard_clip_min = 5
    hard_clip_max = 100
    hd = LogNormal(log2_space_mean, log2_space_std, hard_clip_min=hard_clip_min, hard_clip_max=hard_clip_max)

    samples = hd_rvs_many(hd)

    samples_median = np.median(samples)
    assert 25 < samples_median < 35
    samples_std = np.std(samples)
    assert 20 < samples_std < 30
    assert hd.pdf(0.) == 0.
    assert abs(hd.pdf(6.) - 0.03385080142719004) < 1e-6
    assert abs(hd.pdf(10.) - 0.024999599033243936) < 1e-6
    assert hd.cdf(0.) == 0.
    assert abs(hd.cdf(6.) - 0.03560663481768936) < 1e-6
    assert abs(hd.cdf(10.) - 0.15112888563249155) < 1e-6

    assert hd.min() == hard_clip_min
    assert hd.max() == hard_clip_max
    assert abs(hd.mean() - 38.110960930274594) < 1e-6
    assert abs(hd.var() - 728.7599668053633) < 1e-6
    assert abs(hd.std() - 26.995554574880718) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs((hd.mean() - np.mean(samples)) / hd.mean()) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 1e-1


def test_gaussian_distribution_mixture():
    distribution_amplitudes = [1, 1, 1]
    means = [-2, 0, 2]
    stds = [1, 1, 1]
    distribution_mins = [None for _ in range(len(means))]
    distribution_max = [None for _ in range(len(means))]

    hd = DistributionMixture.build_gaussian_mixture(distribution_amplitudes, means, stds, distribution_mins,
                                                    distribution_max)

    samples = hd_rvs_many(hd)

    samples_median = np.median(samples)
    assert -0.5 < samples_median < 0.5
    samples_std = np.std(samples)
    assert 1 < samples_std < 4
    assert abs(hd.pdf(-2.) - 0.1510223590467952) < 1e-6
    assert abs(hd.pdf(0.) - 0.16897473780926958) < 1e-6
    assert abs(hd.pdf(2.) - 0.1510223590467952) < 1e-6
    assert abs(hd.cdf(-2.) - 0.17426060106333743) < 1e-6
    assert abs(hd.cdf(0.) - 0.5) < 1e-6
    assert abs(hd.cdf(2.) - 0.8257393989366625) < 1e-6

    assert hd.min() == -np.inf
    assert hd.max() == np.inf
    assert abs(hd.mean() - 0.0) < 1e-6
    assert abs(hd.var() - 11. / 3) < 1e-6
    assert abs(hd.std() - math.sqrt(11. / 3)) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs(hd.var() - np.var(samples)) < 1e-1


def test_gaussian_distribution_mixture_truncated():
    distribution_amplitudes = [1, 1, 1]
    means = [-2, 0, 2]
    stds = [1, 1, 1]
    distribution_mins = [-2.5, -0.5, 1.5]
    distribution_max = [-1.5, 0.5, 2.5]

    hd = DistributionMixture.build_gaussian_mixture(distribution_amplitudes, means, stds, distribution_mins,
                                                    distribution_max)

    samples = hd_rvs_many(hd)

    assert np.all(np.logical_and(np.array(samples) >= distribution_mins[0], np.array(samples) <= distribution_max[0]) |
                  np.logical_and(np.array(samples) >= distribution_mins[1], np.array(samples) <= distribution_max[1]) |
                  np.logical_and(np.array(samples) >= distribution_mins[2], np.array(samples) <= distribution_max[2]))
    samples_median = np.median(samples)
    assert -0.5 < samples_median < 0.5
    samples_std = np.std(samples)
    assert 1 < samples_std < 4
    assert abs(hd.pdf(-2.) - 0.3472763257323177) < 1e-6
    assert abs(hd.pdf(0.) - 0.3472763257323177) < 1e-6
    assert abs(hd.pdf(2.) - 0.3472763257323177) < 1e-6
    assert abs(hd.cdf(-2.) - 0.16666666666666666) < 1e-6
    assert abs(hd.cdf(0.) - 0.5) < 1e-6
    assert abs(hd.cdf(2.) - 0.8333333333333333) < 1e-6

    assert hd.min() == min(distribution_mins)
    assert hd.max() == max(distribution_max)
    assert abs(hd.mean() - 0.0) < 1e-6
    assert abs(hd.var() - 2.747255821267478) < 1e-6
    assert abs(hd.std() - 1.6574847876428545) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs(hd.var() - np.var(samples)) < 1e-1


def test_gaussian_distribution_mixture_log():
    distribution_amplitudes = [1, 1, 1]
    means = [1, 2, 3]
    stds = [1, 1, 1]
    distribution_mins = [None for _ in range(len(means))]
    distribution_max = [None for _ in range(len(means))]

    hd = DistributionMixture.build_gaussian_mixture(
        distribution_amplitudes, means, stds,
        distribution_mins, distribution_max, use_logs=True)

    samples = hd_rvs_many(hd)

    samples_median = np.median(samples)
    assert 1.5 < samples_median < 2.5
    samples_std = np.std(samples)
    assert 1.5 < hd.mean() < 2.5
    assert 1 < samples_std < 4
    assert 1 < hd.std() < 2
    assert abs(hd.pdf(-2.) - 0.) == 0
    assert abs(hd.pdf(1) - 1.0 / 3.0) < 0.1
    assert abs(hd.pdf(2) - 1.0 / 3.0) < 0.1
    assert abs(hd.pdf(3) - 1.0 / 3.0) < 0.1
    assert abs(hd.cdf(1) - 1.0 / 4.0) < 0.1
    assert abs(hd.cdf(2) - 2.0 / 4.0) < 0.1
    assert abs(hd.cdf(3) - 3.0 / 4.0) < 0.1
    assert hd.pdf(5.) < 0.05
    assert hd.min() == 0
    assert hd.max() == np.inf
    assert abs(hd.mean() - 2.15) < 0.015
    assert abs(hd.std() ** 2 - hd.var()) < 0.001
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 0.05
    assert abs(hd.var() - np.var(samples)) < 0.15


def test_gaussian_distribution_mixture_quantized():
    distribution_amplitudes = [1, 1, 1]
    means = [-2, 0, 2]
    stds = [1, 1, 1]
    distribution_mins = [None for _ in range(len(means))]
    distribution_max = [None for _ in range(len(means))]

    hd = DistributionMixture.build_gaussian_mixture(
        distribution_amplitudes, means, stds, distribution_mins, distribution_max,
        use_quantized_distributions=True
    )

    samples = hd_rvs_many(hd)

    samples_median = np.median(samples)
    assert -0.5 < samples_median < 0.5
    samples_std = np.std(samples)
    assert 1 < samples_std < 4
    assert abs(hd.pdf(-2.) - 0.147917229965673) < 1e-6
    assert abs(hd.pdf(-1.5) - 0.) < 1e-6
    assert abs(hd.pdf(1.) - 0.16314590372033272) < 1e-6
    assert abs(hd.pdf(5.) - 0.001993471656810324) < 1e-6
    assert abs(hd.cdf(-2.) - 0.2528340972073022) < 1e-6
    assert abs(hd.cdf(-1.5) - 0.2528340972073022) < 1e-6
    assert abs(hd.cdf(1.) - 0.7471659027926978) < 1e-6
    assert abs(hd.cdf(5.) - 0.99992245064379) < 1e-6

    assert hd.min() == -np.inf
    assert hd.max() == np.inf
    assert abs(hd.mean() - 0.) < 1e-6
    assert abs(hd.var() - 3.749999989027785) < 1e-6
    assert abs(hd.std() - 1.9364916702706947) < 1e-6
    # Verify that hd mean and variance also correspond to mean and variance of sampling.
    assert abs(hd.mean() - np.mean(samples)) < 1e-1
    assert abs((hd.var() - np.var(samples)) / hd.var()) < 1e-1



import copy
from typing import Type

import pytest
from neuraxle.base import TrialStatus, synchroneous_flow_method
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT,
                                           RETRAIN_TRIAL_SPLIT_ID,
                                           BaseDataclass, ClientDataclass,
                                           MetricResultsDataclass,
                                           ProjectDataclass, RootDataclass,
                                           RoundDataclass, ScopedLocation,
                                           TrialDataclass, TrialSplitDataclass,
                                           as_named_odict, dataclass_2_id_attr,
                                           from_json, to_json)

SOME_METRIC_NAME = 'MAE'
HYPERPARAMS = {'learning_rate': 0.01}


SOME_METRIC_RESULTS_DATACLASS = MetricResultsDataclass(
    metric_name=SOME_METRIC_NAME,
    validation_values=[3, 2, 1],
    train_values=[2, 1, 0],
    higher_score_is_better=False,
)
SOME_TRIAL_DATACLASS = TrialDataclass(
    trial_number=0,
    hyperparams=HYPERPARAMS,
).start()
SOME_TRIAL_SPLIT_DATACLASS = TrialSplitDataclass(
    split_number=0,
    metric_results=as_named_odict(SOME_METRIC_RESULTS_DATACLASS),
    hyperparams=HYPERPARAMS,
).start()
SOME_TRIAL_SPLIT_DATACLASS.end(TrialStatus.SUCCESS)
SOME_TRIAL_DATACLASS.store(SOME_TRIAL_SPLIT_DATACLASS)
SOME_TRIAL_DATACLASS.end(TrialStatus.SUCCESS)
SOME_ROUND_DATACLASS = RoundDataclass(
    round_number=0,
    trials=[SOME_TRIAL_DATACLASS],
    main_metric_name=SOME_METRIC_NAME,
)
SOME_CLIENT_DATACLASS = ClientDataclass(
    client_name=DEFAULT_CLIENT,
    rounds=[SOME_ROUND_DATACLASS],
)
SOME_PROJECT_DATACLASS = ProjectDataclass(
    project_name=DEFAULT_PROJECT,
    clients=as_named_odict(SOME_CLIENT_DATACLASS),
)
SOME_ROOT_DATACLASS = RootDataclass(
    projects=as_named_odict(SOME_PROJECT_DATACLASS),
)

SOME_FULL_SCOPED_LOCATION: ScopedLocation = ScopedLocation(
    DEFAULT_PROJECT, DEFAULT_CLIENT, 0, 0, 0, SOME_METRIC_NAME
)

ALL_DATACLASSES = [
    SOME_ROOT_DATACLASS,
    SOME_PROJECT_DATACLASS,
    SOME_CLIENT_DATACLASS,
    SOME_ROUND_DATACLASS,
    SOME_TRIAL_DATACLASS,
    SOME_TRIAL_SPLIT_DATACLASS,
    SOME_METRIC_RESULTS_DATACLASS,
]


@pytest.mark.parametrize("scope_slice_len", list(range(len(ALL_DATACLASSES))))
def test_dataclass_getters(scope_slice_len: int):
    expected_dataclass: BaseDataclass = ALL_DATACLASSES[scope_slice_len]
    dataclass_type: Type[BaseDataclass] = expected_dataclass.__class__
    sliced_scope = SOME_FULL_SCOPED_LOCATION[:dataclass_type]
    assert len(sliced_scope) == scope_slice_len

    sliced_scope = SOME_FULL_SCOPED_LOCATION[:scope_slice_len]
    assert len(sliced_scope) == scope_slice_len

    dc = SOME_ROOT_DATACLASS[sliced_scope]

    assert isinstance(dc, dataclass_type)
    assert dc.get_id() == expected_dataclass.get_id()
    assert dc.get_id() == sliced_scope[scope_slice_len - 1]
    assert dc.get_id() == SOME_FULL_SCOPED_LOCATION[dataclass_type]
    assert dc == expected_dataclass


@pytest.mark.parametrize("dataclass_type, scope", [
    (ProjectDataclass, ScopedLocation(DEFAULT_PROJECT)),
    (ClientDataclass, ScopedLocation(DEFAULT_PROJECT, DEFAULT_CLIENT)),
])
def test_base_empty_default_dataclass_getters(
    dataclass_type: Type[BaseDataclass],
    scope: ScopedLocation,
):
    root: RootDataclass = RootDataclass()

    dc = root[scope]

    assert isinstance(dc, dataclass_type)
    assert dc.get_id() == scope.as_list()[-1]


@pytest.mark.parametrize("cp", [copy.copy, copy.deepcopy])
def test_scoped_location_can_copy_and_change(cp):
    sl = ScopedLocation(DEFAULT_PROJECT, DEFAULT_CLIENT, 0, 0, 0, SOME_METRIC_NAME)

    sl_copy = cp(sl)
    sl_copy.pop()
    sl_copy.pop()

    assert sl_copy != sl
    assert len(sl_copy) == len(sl) - 2


@pytest.mark.parametrize("dataclass_type", list(dataclass_2_id_attr.keys()))
def test_dataclass_id_attr_get_set(dataclass_type):
    _id = 9000
    dc = dataclass_type().set_id(_id)

    assert dc.get_id() == _id


def test_dataclass_from_dict_to_dict():
    root: RootDataclass = SOME_ROOT_DATACLASS

    root_as_dict = root.to_dict()
    root_restored = RootDataclass.from_dict(root_as_dict)

    assert SOME_METRIC_RESULTS_DATACLASS == root_restored[SOME_FULL_SCOPED_LOCATION]
    assert root == root_restored


def test_dataclass_from_json_to_json():
    root: RootDataclass = SOME_ROOT_DATACLASS

    root_as_dict = root.to_dict()
    root_as_json = to_json(root_as_dict)
    root_restored_dc = from_json(root_as_json)

    assert SOME_METRIC_RESULTS_DATACLASS == root_restored_dc[SOME_FULL_SCOPED_LOCATION]
    assert root == root_restored_dc


def test_trial_dataclass_can_store_and_contains_retrain_split():
    tc: TrialDataclass = TrialDataclass(trial_number=5)
    tc.store(TrialSplitDataclass(split_number=0))
    tc.store(TrialSplitDataclass(split_number=1))
    tc.store(TrialSplitDataclass(split_number=2))
    sl: ScopedLocation = ScopedLocation.default(
        round_number=0, trial_number=5, split_number=RETRAIN_TRIAL_SPLIT_ID)
    tsc: TrialSplitDataclass = TrialSplitDataclass(split_number=RETRAIN_TRIAL_SPLIT_ID)

    tc.store(tsc)

    assert len(tc) == 3
    assert sl in tc
    assert tsc.get_id() == RETRAIN_TRIAL_SPLIT_ID
    assert tc.retrained_split == tsc
    assert tc[sl] == tc.retrained_split



from typing import List, Iterable

import numpy as np

from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ExpandDim
from neuraxle.steps.misc import HandleCallbackStep, TapeCallbackFunction


def test_expand_dim_transform():
    di = np.array(range(10))
    eo = [None] * 10
    handle_fit_callback = TapeCallbackFunction()
    handle_transform_callback = TapeCallbackFunction()
    handle_fit_transform_callback = TapeCallbackFunction()
    p = Pipeline([
        ExpandDim(
            HandleCallbackStep(
                handle_fit_callback,
                handle_transform_callback,
                handle_fit_transform_callback
            )
        )
    ])

    outputs = p.transform(di)

    assert np.array_equal(outputs, di)
    assert handle_fit_callback.data == []
    assert np.array_equal(
        np.array(handle_transform_callback.data[0][0].di),
        np.array([di])
    )
    assert np.array_equal(
        np.array(handle_transform_callback.data[0][0].eo),
        np.array([eo])
    )
    assert handle_fit_transform_callback.data == []


def test_expand_dim_fit():
    handle_fit_callback = TapeCallbackFunction()
    handle_transform_callback = TapeCallbackFunction()
    handle_fit_transform_callback = TapeCallbackFunction()
    p = Pipeline([
        ExpandDim(
            HandleCallbackStep(
                handle_fit_callback,
                handle_transform_callback,
                handle_fit_transform_callback
            )
        )
    ])

    p = p.fit(np.array(range(10)), np.array(range(10)))

    assert handle_transform_callback.data == []
    assert handle_fit_transform_callback.data == []
    assert np.array_equal(
        np.array(handle_fit_callback.data[0][0].data_inputs),
        np.array([np.array(range(10))])
    )
    assert np.array_equal(
        np.array(handle_fit_callback.data[0][0].expected_outputs),
        np.array([np.array(range(10))])
    )


def test_expand_dim_fit_transform():
    handle_fit_callback = TapeCallbackFunction()
    handle_transform_callback = TapeCallbackFunction()
    handle_fit_transform_callback = TapeCallbackFunction()
    p = Pipeline([
        ExpandDim(
            HandleCallbackStep(
                handle_fit_callback,
                handle_transform_callback,
                handle_fit_transform_callback
            )
        )
    ])

    p, outputs = p.fit_transform(np.array(range(10)), np.array(range(10)))

    assert np.array_equal(outputs, np.array(range(10)))
    assert handle_transform_callback.data == []
    assert handle_fit_callback.data == []
    assert np.array_equal(
        np.array(handle_fit_transform_callback.data[0][0].data_inputs),
        np.array([np.array(range(10))])
    )
    assert np.array_equal(
        np.array(handle_fit_transform_callback.data[0][0].expected_outputs),
        np.array([np.array(range(10))])
    )



from neuraxle.base import Identity
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN
from neuraxle.steps.output_handlers import OutputTransformerWrapper


def test_apply_on_pipeline_with_positional_argument_should_call_method_on_each_steps():
    pipeline = Pipeline([MultiplyByN(1), MultiplyByN(1)])

    pipeline.apply('_set_hyperparams', hyperparams=HyperparameterSamples({
        'multiply_by': 2,
        'MultiplyByN__multiply_by': 3,
        'MultiplyByN1__multiply_by': 4
    }))

    assert pipeline.get_hyperparams()['multiply_by'] == 2
    assert pipeline['MultiplyByN'].get_hyperparams()['multiply_by'] == 3
    assert pipeline['MultiplyByN1'].get_hyperparams()['multiply_by'] == 4


def test_apply_method_on_pipeline_should_call_method_on_each_steps():
    pipeline = Pipeline([MultiplyByN(1), MultiplyByN(1)])

    pipeline.apply(
        lambda step: step._set_hyperparams(HyperparameterSamples({'multiply_by': 2}))
    )

    assert pipeline.get_hyperparams()['multiply_by'] == 2
    assert pipeline['MultiplyByN'].get_hyperparams()['multiply_by'] == 2
    assert pipeline['MultiplyByN1'].get_hyperparams()['multiply_by'] == 2


def test_apply_on_pipeline_with_meta_step_and_positional_argument():
    pipeline = Pipeline([OutputTransformerWrapper(MultiplyByN(1)), MultiplyByN(1)])

    pipeline.apply('_set_hyperparams', hyperparams=HyperparameterSamples({
        'multiply_by': 2,
        'OutputTransformerWrapper__multiply_by': 3,
        'OutputTransformerWrapper__MultiplyByN__multiply_by': 4,
        'MultiplyByN__multiply_by': 5
    }))

    assert pipeline.get_hyperparams()['multiply_by'] == 2
    assert pipeline['OutputTransformerWrapper'].get_hyperparams()['multiply_by'] == 3
    assert pipeline['OutputTransformerWrapper'].wrapped.get_hyperparams()['multiply_by'] == 4
    assert pipeline['MultiplyByN'].get_hyperparams()['multiply_by'] == 5


def test_apply_method_on_pipeline_with_meta_step_should_call_method_on_each_steps():
    pipeline = Pipeline([OutputTransformerWrapper(MultiplyByN(1)), MultiplyByN(1)])

    pipeline.apply(
        lambda step: step._set_hyperparams(HyperparameterSamples({'multiply_by': 2}))
    )

    assert pipeline.get_hyperparams()['multiply_by'] == 2
    assert pipeline['OutputTransformerWrapper'].get_hyperparams()['multiply_by'] == 2
    assert pipeline['OutputTransformerWrapper'].wrapped.get_hyperparams()['multiply_by'] == 2
    assert pipeline['MultiplyByN'].get_hyperparams()['multiply_by'] == 2


def test_has_children_mixin_apply_should_apply_method_to_direct_childrends():
    p = Pipeline([
        ('a', Identity()),
        ('b', Identity()),
        Pipeline([
            ('c', Identity()),
            ('d', Identity())
        ]),
    ])

    p.apply('_set_hyperparams', ra=None, hyperparams=HyperparameterSamples({
        'a__hp': 0,
        'b__hp': 1,
        'Pipeline__hp': 2
    }))

    assert p['a'].hyperparams.to_flat_dict()['hp'] == 0
    assert p['b'].hyperparams.to_flat_dict()['hp'] == 1
    assert p['Pipeline'].hyperparams.to_flat_dict()['hp'] == 2


def test_has_children_mixin_apply_should_apply_method_to_recursive_childrends():
    p = Pipeline([
        ('a', Identity()),
        ('b', Identity()),
        Pipeline([
            ('c', Identity()),
            ('d', Identity())
        ]),
    ])

    p.apply('_set_hyperparams', ra=None, hyperparams=HyperparameterSamples({
        'Pipeline__c__hp': 3,
        'Pipeline__d__hp': 4
    }))

    assert p['Pipeline']['c'].hyperparams.to_flat_dict()['hp'] == 3
    assert p['Pipeline']['d'].hyperparams.to_flat_dict()['hp'] == 4


def test_has_children_mixin_apply_should_return_recursive_dict_to_direct_childrends():
    p = Pipeline([
        ('a', Identity().set_hyperparams(HyperparameterSamples({'hp': 0}))),
        ('b', Identity().set_hyperparams(HyperparameterSamples({'hp': 1})))
    ])

    results = p.apply('_get_hyperparams', ra=None)

    assert results.to_flat_dict()['a__hp'] == 0
    assert results.to_flat_dict()['b__hp'] == 1


def test_has_children_mixin_apply_should_return_recursive_dict_to_recursive_childrends():
    p = Pipeline([
        Pipeline([
            ('c', Identity().set_hyperparams(HyperparameterSamples({'hp': 3}))),
            ('d', Identity().set_hyperparams(HyperparameterSamples({'hp': 4})))
        ]).set_hyperparams(HyperparameterSamples({'hp': 2})),
    ])

    results = p.apply('_get_hyperparams', ra=None)

    assert results['Pipeline__hp'] == 2
    assert results['Pipeline__c__hp'] == 3
    assert results['Pipeline__d__hp'] == 4



"""
Tree parzen estimator
====================================
Code for tree parzen estimator auto ml.
"""
from collections import Counter
from operator import itemgetter
from typing import Any, List, Tuple, Union

import numpy as np
from neuraxle.hyperparams.distributions import (
    Choice, DiscreteHyperparameterDistribution, DistributionMixture,
    HPSampledValue, HyperparameterDistribution, LogSpaceDistributionMixin,
    PriorityChoice, Quantized)
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.metaopt.data.aggregates import Round, Trial
from neuraxle.metaopt.data.vanilla import BaseHyperparameterOptimizer
from neuraxle.metaopt.validation import GridExplorationSampler

_QUANTIZED_DISTRIBUTION = (Quantized,)


class TreeParzenEstimator(BaseHyperparameterOptimizer):
    """
    This is a Tree Parzen Estimator (TPE) algorithm as found in Hyperopt,
    that is better than the Random Search algorithm and supports supporting
    intelligent exploration v.s. exploitation of the search space over time,
    using Neuraxle hyperparameters.

    Here, the algorithm is modified compared to the original one, as it uses a
    :class:`neuraxle.metaopt.automl.GridExplorationSampler` instead of a random search
    to pick the first exploration samples to furthermore explore the space at the beginning.
    """

    def __init__(
            self,
            number_of_initial_random_step: int = 15,
            quantile_threshold: float = 0.3,
            number_good_trials_max_cap: int = 25,
            number_possible_hyperparams_candidates: int = 100,
            use_linear_forgetting_weights: bool = False,
            number_recent_trials_at_full_weights: int = 25
    ):
        """
        Initialize the TPE with some configuration.

        :param number_of_initial_random_step: Number of random steps to take before starting the optimization.
        :param quantile_threshold: threshold between 0 and 1 representing the proportion of good trials to keep.
        :param number_good_trials_max_cap: maximum number of good trials to keep, that will cap the `quantile_threshold`.
        :param number_possible_hyperparams_candidates: number of possible hyperparams candidates to explore in the `good / bad` trials posterior.
        :param use_linear_forgetting_weights: if True, the weights will be linearly decreasing past the `number_recent_trial_at_full_weights`.
        :param number_recent_trials_at_full_weights: number of recent trials to use at full weights before linear forgetting when `use_linear_forgetting_weights` is set to True.
        """
        super().__init__()

        self.number_of_initial_random_step: int = number_of_initial_random_step
        self.number_possible_hyperparams_candidates: int = number_possible_hyperparams_candidates

        self.mixture_factory: _DividedMixturesFactory = _DividedMixturesFactory(
            quantile_threshold,
            number_good_trials_max_cap,
            use_linear_forgetting_weights,
            number_recent_trials_at_full_weights,
        )

        self.initial_auto_ml_algo: BaseHyperparameterOptimizer = GridExplorationSampler(
            expected_n_trials=number_of_initial_random_step
        )

    def find_next_best_hyperparams(self, round_scope: Round) -> HyperparameterSamples:
        """
        Find the next best hyperparams using previous trials.

        :param round_scope: round scope
        :return: next best hyperparams
        """

        # Perform a first pseudo-randomized search:
        if len(round_scope) < self.number_of_initial_random_step:
            return self.initial_auto_ml_algo.find_next_best_hyperparams(round_scope)

        # Create gaussian mixture of good and gaussian mixture of bads. Lists here are on a per-hp basis:
        hyperparams_keys, divided_good_and_bad_distrs = self.mixture_factory.create_from(round_scope)

        # Sample the next hyperparams finally:
        return self._sample_next_hyperparams_from_gaussians_div(
            hyperparams_keys, divided_good_and_bad_distrs)

    def _sample_next_hyperparams_from_gaussians_div(
        self,
        hp_keys: List[str],
        divided_good_and_bad_distrs: List['_DividedTPEPosteriors']
    ) -> HyperparameterSamples:
        best_hyperparams: List[Tuple[str, HPSampledValue]] = []

        for (hp_key, good_bad_posterior_div) in zip(hp_keys, divided_good_and_bad_distrs):

            best_next_hyperparam_value: HPSampledValue = self._sample_one_hp_from_gaussians_div(
                hp_key, good_bad_posterior_div)

            best_hyperparams.append((hp_key, best_next_hyperparam_value))

        return HyperparameterSamples(best_hyperparams)

    def _sample_one_hp_from_gaussians_div(
        self,
        hp_key: List[str],
        good_bad_posterior_div: '_DividedTPEPosteriors'
    ) -> HPSampledValue:

        best_next_hyperparam_value: HPSampledValue = None
        best_ratio: float = None
        for _ in range(self.number_possible_hyperparams_candidates):

            possible_new_hyperparm, proba_ratio = good_bad_posterior_div.rvs_good_with_pdf_division_proba()

            if best_next_hyperparam_value is None or proba_ratio > best_ratio:
                best_next_hyperparam_value = possible_new_hyperparm
                best_ratio = proba_ratio

        return best_next_hyperparam_value


class _DividedMixturesFactory:

    def __init__(
        self,
        quantile_threshold: float,
        number_good_trials_max_cap: int,
        use_linear_forgetting_weights: bool,
        number_recent_trials_at_full_weights: int,
    ):
        self.quantile_threshold: float = quantile_threshold
        self.number_good_trials_max_cap: int = number_good_trials_max_cap
        self.use_linear_forgetting_weights = use_linear_forgetting_weights
        self.number_recent_trials_at_full_weights = number_recent_trials_at_full_weights

    def create_from(
        self,
        round_scope: Round,
    ) -> Tuple[List[str], List['_DividedTPEPosteriors']]:

        # Split trials into good and bad using quantile threshold.
        good_trials: List[Trial] = []
        bad_trials: List[Trial] = []
        good_trials, bad_trials = self._split_good_and_bad_trials(round_scope)

        flat_hp_space_tuples: List[(str, HyperparameterDistribution)] = list(
            round_scope.hp_space.to_flat_dict().items())
        hyperparams_keys: List[str] = list(map(itemgetter(0), flat_hp_space_tuples))

        good_posteriors: List[HyperparameterDistribution] = self._create_posterior(
            flat_hp_space_tuples, good_trials)
        bad_posteriors: List[HyperparameterDistribution] = self._create_posterior(
            flat_hp_space_tuples, bad_trials)

        divided_good_and_bad_distrs: _DividedMixturesFactory = [
            _DividedTPEPosteriors(g, b)
            for (g, b)
            in zip(good_posteriors, bad_posteriors)
        ]

        return hyperparams_keys, divided_good_and_bad_distrs

    def _split_good_and_bad_trials(self, round_scope: Round) -> Tuple[List[Trial], List[Trial]]:

        # Split trials into good and bad using quantile threshold.
        successful_trials: List[Trial] = [
            trial for trial in round_scope._trials
            if trial.is_success()
        ]
        trials_scores: List[Tuple[float, Trial]] = [t.get_avg_validation_score() for t in successful_trials]

        trial_sorted_indexes: List[int] = np.argsort(trials_scores)
        if round_scope.is_higher_score_better():
            trial_sorted_indexes = list(reversed(trial_sorted_indexes))

        # In hyperopt they use this to split, where default_gamma_cap = 25. They clip the max of item they use in the good item.
        # default_gamma_cap is link to the number of recent_trial_at_full_weight also.
        # n_below = min(int(np.ceil(gamma * np.sqrt(len(l_vals)))), gamma_cap)
        n_good = int(min(
            self.number_good_trials_max_cap,
            np.ceil(self.quantile_threshold * len(trials_scores)),
        ))

        good_trials_indexes = trial_sorted_indexes[:n_good]
        bad_trials_indexes = trial_sorted_indexes[n_good:]

        good_trials: List[Trial] = [successful_trials[i] for i in good_trials_indexes]
        bad_trials: List[Trial] = [successful_trials[i] for i in bad_trials_indexes]

        return good_trials, bad_trials

    def _create_posterior(
        self,
        flat_hp_space_tuples: List[Tuple[str, HyperparameterDistribution]],
        trials: List[Trial],
    ) -> List[HyperparameterDistribution]:

        # Loop through all hyperparams
        posterior_distributions: List[HyperparameterDistribution] = []
        for (hp_key, hp_distribution) in flat_hp_space_tuples:
            # Typing:
            hp_key: str = hp_key
            hp_distribution: HyperparameterDistribution = hp_distribution

            # Get these trials' hyperparam values for this hyperparam
            trial_hyperparams: List[HPSampledValue] = [trial.get_hyperparams()[hp_key] for trial in trials]

            if hp_distribution.is_discrete():
                posterior_distribution: Union[Choice, PriorityChoice] = self._reweights_categorical(
                    hp_distribution, trial_hyperparams)
            else:
                posterior_distribution: DistributionMixture = self._create_gaussian_mixture(
                    hp_distribution, trial_hyperparams)

            posterior_distributions.append(posterior_distribution)

        return posterior_distributions

    def _reweights_categorical(
        self,
        discrete_distribution: DiscreteHyperparameterDistribution,
        trial_hyperparameters: List[HPSampledValue]
    ) -> Union[Choice, PriorityChoice]:
        # For discrete categorical distribution
        # We need to reweights probability depending on trial counts.
        probas: List[float] = discrete_distribution.probabilities()
        values: List[Any] = discrete_distribution.values()

        number_probas = len(probas)

        # Since the reweighted is proportional to N*p_i + C_i,
        # where N is the number of probas, p_i is the original probas and c_i is the count.
        reweighted_probas = np.array(number_probas) * probas

        # Count number of occurence for each values.
        count_trials = Counter(trial_hyperparameters)
        values_keys = list(count_trials.keys())
        counts = list(count_trials.values())

        for value_key, count in zip(values_keys, counts):
            # find index in the orignal probas and values.
            index_value = values.index(value_key)

            # Calculate reweigthed proba.
            reweighted_probas[index_value] += count

        # Normalize reweighted probas
        reweighted_probas = np.array(reweighted_probas)
        reweighted_probas = reweighted_probas / np.sum(reweighted_probas)

        if isinstance(discrete_distribution, PriorityChoice):
            return PriorityChoice(values, probas=reweighted_probas)

        return Choice(values, probas=reweighted_probas)

    def _create_gaussian_mixture(
            self,
            continuous_distribution: HyperparameterDistribution,
            trial_hyperparameters: List[HPSampledValue]
    ) -> DistributionMixture:

        use_logs = False
        if isinstance(continuous_distribution, LogSpaceDistributionMixin):
            use_logs = True

        use_quantized_distributions = False
        if isinstance(continuous_distribution, Quantized):
            use_quantized_distributions = True
            if isinstance(continuous_distribution.hd, LogSpaceDistributionMixin):
                use_logs = True

        # Find means, std, amplitudes, min and max.
        distribution_amplitudes, means, stds, distributions_mins, distributions_max = self._adaptive_parzen_normal(
            continuous_distribution,
            trial_hyperparameters)

        # Create appropriate gaussian mixture that wrapped all hyperparams.
        gmm: DistributionMixture = DistributionMixture.build_gaussian_mixture(
            distribution_amplitudes=distribution_amplitudes,
            means=means,
            stds=stds,
            distributions_mins=distributions_mins,
            distributions_max=distributions_max,
            use_logs=use_logs,
            use_quantized_distributions=use_quantized_distributions
        )

        return gmm

    def _adaptive_parzen_normal(self, hyperparam_distribution: HyperparameterDistribution, distribution_trials: List[HPSampledValue]):

        hp_std = hyperparam_distribution.std()

        means = np.array(distribution_trials)
        distributions_mins = hyperparam_distribution.min() * np.ones_like(means)
        distributions_max = hyperparam_distribution.max() * np.ones_like(means)

        means, stds = self._compute_distributions_means_stds(hyperparam_distribution, distribution_trials)

        # Magic stds formula from hyperopt:
        min_std = hp_std / min(100.0, (1.0 + len(means)))
        max_std = hp_std / 1.0
        stds = np.clip(stds, min_std, max_std)

        distribution_amplitudes: np.ndarray = self._generate_linear_forget_weights(len(means))

        return distribution_amplitudes, means, stds, distributions_mins, distributions_max

    def _compute_distributions_means_stds(self, hyperparam_distribution: HyperparameterDistribution, distribution_trials: List[HPSampledValue]):

        prior_mean = hyperparam_distribution.mean()
        prior_sigma = hyperparam_distribution.std()

        # Initialize with prior mean and sigma and replace when appropriate.
        means = np.array(sorted(distribution_trials)) if len(distribution_trials) > 0 else np.array([prior_mean])
        stds = np.array([prior_sigma])

        if len(means) > 1:
            # Compute the STDs specially with a far-left and far-right tail.
            stds = np.zeros_like(means)
            # Element-wise std computation is from the maximal difference in closest neighbor.
            stds[1:-1] = np.maximum(means[1:-1] - means[:-2],
                                    means[2:] - means[1:-1])
            left_std = means[1] - means[0]
            right_std = means[-1] - means[-2]
            stds[0] = left_std
            stds[-1] = right_std

        return means, stds

    def _generate_linear_forget_weights(self, number_samples: int) -> np.ndarray:
        # From the TPE article: TPE substitutes an equally-weighted mixture of that prior with Gaussians centered at each observations.

        if self.use_linear_forgetting_weights and number_samples >= self.number_recent_trials_at_full_weights:
            weights_ramp = np.linspace(
                start=1.0 / number_samples,
                stop=1.0,
                num=number_samples - self.number_recent_trials_at_full_weights
            )
            weights_flat = np.ones(self.number_recent_trials_at_full_weights)

            distribution_amplitudes = np.concatenate((weights_ramp, weights_flat), axis=0)
        else:
            distribution_amplitudes = np.ones(number_samples)

        return distribution_amplitudes


class _DividedTPEPosteriors:
    """
    Sample possible new hyperparams in the `good_trials`.

    Verify if we use the ratio directly or we use the loglikelihood of b_post under both distribution like hyperopt.
    In hyperopt, they use the following to calculate the log likelihood of b_post under both distributions:

    .. code-block:: python
        below_llik = fn_lpdf(*([b_post] + b_post.pos_args), **b_kwargs)
        above_llik = fn_lpdf(*([b_post] + a_post.pos_args), **a_kwargs)
        improvement = below_llik - above_llik
        new_node = scope.broadcast_best(b_post, improvement)
        new_node = scope.broadcast_best(b_post, below_llik, above_llik)

    Verify ratio good pdf versus bad pdf for all possible new hyperparams.

    The gist of it is that this is as dividing good probabilities by bad probabilities.

    We used what is described in the article which is the ratio (gamma + g(x) / l(x) ( 1- gamma))^-1 that we have to maximize.

    Since there is `^-1`, we have to maximize `l(x) / g(x)`.

    Only the best ratio is kept and is the new best hyperparams.

    They seem to take log of pdf and not pdf directly so it's probable to have `-` instead of `/` as an implementation. Regardless, both operators are transitive in nature for `max` comparisons in their regular or log space.

    # TODO: Maybe they use the likelyhood to sum over all possible parameters to find the max so it become a join distribution of all hyperparameters, would make sense.

    # TODO: verify is for quantized we do not want to do cdf(value higher) - cdf(value lower) to have pdf.
    """

    def __init__(self, good_trials: HyperparameterDistribution, bad_trials: HyperparameterDistribution):
        self.good_trials = good_trials
        self.bad_trials = bad_trials

    def rvs_good_with_pdf_division_proba(self) -> Tuple[HPSampledValue, float]:
        """
        Sample an hyperparameter from the good distribution and return the probability ratio.

        :return: sampled good hyperparameter and probability ratio
        """
        good_hyperparam = self.rvs_good()
        proba_ratio = self.proba_ratio(good_hyperparam)
        return good_hyperparam, proba_ratio

    def rvs_good(self) -> HPSampledValue:
        """
        Sample an hyperparameter from the good distribution.

        :return: sampled good hyperparameter
        """
        return self.good_trials.rvs()

    def proba_ratio(self, possible_new_hyperparm: HPSampledValue) -> float:
        """
        Return the probability ratio of the sampled good hyperparam.

        :return: probability ratio
        """
        proba_ratio = self.good_trials.pdf(possible_new_hyperparm) / self.bad_trials.pdf(possible_new_hyperparm)
        return proba_ratio
